﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
  * @file Dijkstra.hpp
  * @brief алгоритм Дейкстры (поиска кратчайшего пути на графе)
  * @author Кувшинов Д.Р. https://evetro.wordpress.com/
  */
#pragma once
#include <vector>
#include <limits>


/// индекс вершины (для краткости часто отождествляется с вершиной)
typedef std::size_t VertexIndex;

/// Алгоритм Дейкстры: находит пути от заданной вершины до всех других вершин
/**
  Алгоритм Дейкстры поиска кратчайших путей из вершины S до всех остальных вершин графа.
  Веса рёбер, заданные матрицей W, должны быть неотрицательными.
  Вход: N -- количество вершин в графе, S -- начальная вершина, W -- матрица весов рёбер.
  Выход: Pr -- массив предшественников: кратчайший путь из S в i приходит из вершины Pr[i],
    D -- массив длин путей, D[i] -- длина пути из S в i.
*/
template<class Previous, class WeightVector, class WeightMatrix>
void dijkstra
  (
    Previous &Pr          ///< набор предшественников, позволяющий восстановить построенный путь, должен иметь размер >= N
  , WeightVector &D       ///< длина кратчайшего пути от S до вершины графа u равна D[u]
  , const WeightMatrix &W ///< длины рёбер
  , VertexIndex S         ///< индекс начальной вершины
  , VertexIndex N         ///< количество вершин в графе
  )
{
  // признак "непройденности" вершин
  std::vector<bool> spare(N, true);
  spare[S] = false;

  // подготовка
  for (VertexIndex i = 0; i < N; ++i)
  {
    D[i] = W(S, i);
    Pr[i] = S;
  }
     
  D[S] = 0; // путь от начальной до начальной всегда нулевой
     

  // раунды минимизации, на каждом раунде
  // выбирается вершина, ближайшая из непройденных
  for (VertexIndex round = 1; round < N; ++round)
  {
    // найдём ближайшую простым перебором
    VertexIndex v = 0;
    auto minD = std::numeric_limits<typename WeightMatrix::value_type>::max();

    for (VertexIndex u = 0; u < N; ++u)
      if (spare[u] && D[u] < minD)
      {
        v = u;
        minD = D[u];
      }
     
    // пометим вершину v как пройденную
    spare[v] = false;

    // обновим D и Pr для непройденных вершин
    // относительно найденной вершины v
    for (VertexIndex u = 0; u < N; ++u)
      if (spare[u])
      {
        const auto weight = D[v] + W(v, u);
        if (weight < D[u])
        {
          D[u] = weight;
          Pr[u] = v;
        }
      }
  }
}
