﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Scene.hpp
 * @brief Классы событий визуализации, реализация DemoPolicy, класс сцены, склеивающий прочие компоненты
 * @author Кувшинов Д.Р. https://evetro.wordpress.com/
 */
#pragma once
#include <deque>
#include <memory>
#include <iostream>

#include "Dijkstra.hpp"     // реализация алгоритма Дейкстры
#include "Astar.hpp"        // реализация алгоритма A*
#include "TerrainGraph.hpp" // средства рисования ландшафта и граф ландшафта


/// интерфейс события визуализации
struct IEvent
{
  virtual ~IEvent() {}
  /// событие может быть применено к объекту Terrain для внесения изменений в его внешний вид
  virtual void apply(Terrain&) = 0;
};


/// событие "вершина помещена в открытое множество" ("открыта")
class OpenedEvent: public IEvent
{
public:

  OpenedEvent(VertexIndex i, float score)
    : i(i), score(score) {}

  /// калибровка оценки (отображает исходное значение в отрезок [0, 1] )
  void adjust(float minscore, float delta)
  {
    (score -= minscore) /= delta;
  }

  /// применение изменяет цвет
  void apply(Terrain &terr) override
  {
    terr.color(i).red = 1.0f;
    terr.color(i).green = 1.0f - score; // should be adjusted
  }

private:
  VertexIndex i;
  float score;
};


/// событие "вершина помещена в закрытое множество" ("закрыта")
class ClosedEvent: public IEvent
{
  VertexIndex i;

public:
  ClosedEvent(VertexIndex i)
    : i(i) {}

  void apply(Terrain &terr) override
  {
    terr.color(i).red = 0.0f;
  }
};


/// событие "вершина включена в восстанавливаемый путь"
class PathEvent: public IEvent
{
  VertexIndex i;

public:
  PathEvent(VertexIndex i): i(i) {}

  void apply(Terrain &terr) override
  {
    terr.color(i) = avo::WHITE;
  }  
};


/// очередь событий (объектов IEvent)
typedef std::deque<std::unique_ptr<IEvent>> EventQueue;

/// реализация политики DemoPolicy для функции astar
class AstarDemo
{
public:

  /// привязать объект очереди и переменные, в которые будут записаны получаемые в процессе оценки расстояний
  AstarDemo(EventQueue &Q, float &minscore, float &maxscore)
    : Q(Q), minscore(minscore), maxscore(maxscore) {}

  /// создать событие "вершина открыта"
  void openVertex(VertexIndex v, float score)
  {
    if (maxscore < score)
      maxscore = score;
    else if (score < minscore)
      minscore = score;
    Q.emplace_back(new OpenedEvent(v, score));
  }

  /// создать событие "вершина закрыта"
  void closeVertex(VertexIndex v) 
  {
    Q.emplace_back(new ClosedEvent(v));
  }

  /// создать событие "вершина добавлена в путь"
  void nextPathElement(VertexIndex v)
  {
    Q.emplace_back(new PathEvent(v));
  }

private:
  EventQueue &Q;
  float &minscore, &maxscore;
};


/// класс сцены, объединяющий все компоненты программы в единое целое
class Scene
{
  static const std::size_t N = 50 ///< количество строк в карте высот
                         , M = 50 ///< количество столбцов в карте высот
                         ;

  const float xStep ///< шаг сетки по оси 0x
            , yStep ///< шаг сетки по оси 0y
            ;

  /// ландшафт
  Terrain terrain;

  /// граф ландшафта
  TerrainGraph G;

  const avo::Color Gcolor     ///< цвет рёбер графа
                 , ApathColor ///< цвет пути, построенного A*
                 , DpathColor ///< цвет пути, построенного алгоритмом Дейкстры
                 ;

  typedef std::vector<VertexIndex> Path;
  Path Apath ///< путь, построенный A*
     , Dpath ///< путь, построенный алгоритмом Дейкстры
     ;

  /// события
  EventQueue Q;

  /// множитель эвристики алгоритма A*
  float factor;
  /// алгоритм
  enum { normalAstar, bidirectionalAstar }
    current_algorithm;

public:

  bool paintG     ///< признак "рисовать граф ландшафта"
     , paintApath ///< признак "рисовать путь A*"
     , paintDpath ///< признак "рисовать путь алгоритма Дейкстры"
     ;

  Scene()
    : paintG(false), paintApath(false), paintDpath(false)
    , xStep(10.0f),  yStep(10.0f)
    , factor(1.0f)
    , Gcolor     (0.7f , 0.5f, 0.0f)
    , ApathColor (0.1f , 0.7f, 1.0f)
    , DpathColor (0.25f, 1.0f, 0.1f)
    , terrain(N, M + 1)
    , current_algorithm(normalAstar)
  {}


  /// размер ландшафта вдоль оси 0x
  float xSize() const
  {
    return xStep * N;
  }

  /// размер ландшафта вдоль оси 0y
  float ySize() const
  {
    return yStep * M;
  }

  
  /// построить случайный ландшафт и пути на нём
  void newTerrainAndPaths()
  {
    // построить карту высот случайного ландшафта
    HeightMap HM;
    makeLandscape(HM, N, M, 20, 4, 8);

    // получить ландшафт и граф ландшафта из карты высот
    terrain.fromHeightMap(HM, xStep, yStep);
    G.fromTerrain(terrain);
    prepareColors(G, terrain);

    // построить пути
    makeDpath();
    makeApath();
  }


  /// отрисовка сцены
  void paint() const
  {
    terrain.paint();
  
    if (paintG)
    {
      Gcolor.select();
      G.paint();
    }
  
    if (paintApath)
    {
      ApathColor.select();
      G.paint(Apath, G.vertices() - 1);
    }

    if (paintDpath)
    {
      DpathColor.select();
      G.paint(Dpath, G.vertices() - 1);
    }
  }

  /// применить следующее событие из очереди
  bool nextEvent()
  {
    if (Q.empty())
      return false;
    Q.front()->apply(terrain);
    Q.pop_front();
    return true;
  }

  /// применить другой множитель эвристики A* (приводит к перепостроению пути A*)
  void changeFactor(float f)
  {
    if (f != factor)
    {
      factor = f;
      prepareColors(G, terrain);
      makeApath();
    }
  }

  void switchAlgorithm()
  {
    switch (current_algorithm)
    {
      case normalAstar: current_algorithm = bidirectionalAstar; break;
      case bidirectionalAstar: current_algorithm = normalAstar; break;
      default: {}
    }

    prepareColors(G, terrain);
    makeApath();
  }

private:

  /// установить цвета треугольников в соответствии с весами рёбер
  void prepareColors(TerrainGraph &graph, Terrain &terr)
  {
    const VertexIndex N = graph.vertices() - 1;
    const float maxprice = graph.maxPrice(N);

    for (VertexIndex i = 0; i <= N; ++i)
    {
      const float ratio = graph.estimatePrice(i, N) / maxprice;
      terr.color(i) = avo::Color(0.7f - 0.3f * ratio, 0.0f, ratio);
    }
  }

  /// подготовить путь для разметки алгоритмом построения пути
  void preparePath(Path &path, const VertexIndex size)
  {
    path.resize(size);
    for (VertexIndex i = 0; i < size; ++i)
      path[i] = i;
  }


  /// вычислить стоимость пути
  double pathWeight(const WeightsForGraph &W, const Path &P, VertexIndex v)
  {
    double weight = 0.0;
    while (true)
    {
      const auto u = P[v];
      if (u == v)
        return weight;
      weight += W(u, v);
      v = u;
    }
  }


  /// построить путь алгоритмом Дейкстры
  void makeDpath()
  {
    preparePath(Dpath, G.vertices());
     
    // построить пути алгоритмом Дейкстры из вершины 0, содержимое D не используется
    std::vector<float> D(G.vertices());
    WeightsForGraph wfg(G);
    dijkstra(Dpath, D, wfg, 0, G.vertices());

    // стоимость пути
    std::cout << "\nDijkstra path = " << pathWeight(wfg, Dpath, G.vertices() - 1) << std::endl;
  }


  /// построить путь алгоритмом A*, отправляя события визуализации в очередь
  void makeApath()
  {
    Q.clear();
    const auto N = G.vertices(), goal = N - 1;
    preparePath(Apath, N);

    HeuristicForGraph pfg(G, factor);
    WeightsForGraph wfg(G);

    float minscore = std::numeric_limits<float>::max()
        , maxscore = 0.0f;

    AstarDemo demo(Q, minscore, maxscore);

    // построить путь из 0 в (G.vertices()-1)
    switch (current_algorithm)
    {
      case normalAstar:
        astar(Apath, wfg, G, 0, goal, N, pfg, demo);
        break;
      case bidirectionalAstar:
      default:
        astar2(Apath, wfg, G, 0, goal, N, pfg, demo);
        break;
    }

    // отобразить оценки стоимости в [0, 1] для событий открытия вершин
    float delta = maxscore - minscore;
    if (delta == 0.0f)
    {
      delta = minscore;
      minscore *= 0.5f;
    }

    for (auto& p : Q)
      if (auto *e = dynamic_cast<OpenedEvent*>(p.get()))
        e->adjust(minscore, delta);

    // события "прохождения пути"
    for (VertexIndex v = goal;;)
    {
      Q.emplace_back(new PathEvent(v));
      if (v == Apath[v])
        break;
      v = Apath[v];
    }

    // стоимость пути
    std::cout << "A* path = " << pathWeight(wfg, Apath, goal) << std::endl;
  }
};
