﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Collision2D
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Симуляция движения и столкновения двух тяжёлых кружков в плоскости.
 * Кружки можно перетаскивать мышью.
 * Нажатие SPACE/ENTER создаёт новые кружки.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cstdlib>  // rand/srand
#include <cmath>    // hypot
#include <iostream>
#include <avo/Timer.hpp>
#include <avo/Color.hpp>  // RGBA-цвет в 128-битном формате (компоненты типа float)
#include <avo/Vec2.hpp>   // двумерный вектор

using avo::Timer;

//
// базовые определения
//

typedef avo::Vec2d Vec2;

/// вывести текстовое представление вектора в поток вывода
std::ostream& operator<<(std::ostream &os, const Vec2 &v)
{
  return os << '(' << v.x << ", " << v.y << ')';
}


/// твёрдый тяжёлый "кружок"
struct Circle
{
  Vec2 p;   ///< текущая позиция
  Vec2 v;   ///< текущая скорость
  double m; ///< масса
  double r; ///< радиус

  bool dragged; ///< флаг "происходит перетаскивание мышью"

  /// базовый конструктор
  explicit Circle
  (
    const Vec2 &pos = Vec2(),
    const Vec2 &vel = Vec2(),
    double mass = 1.0,
    double rad = 10.0
  )
    : p(pos), v(vel), m(mass), r(rad), dragged(false) {}

  /// нарисовать кружок
  void paint() const
  {
    glLoadIdentity();  // установить единичную матрицу в качестве матрицы модели
    glTranslated(p.x, p.y, 0.0); // домножить матрицу модели на матрицу перемещения в (p.x, p.y)
    glScaled(r, r, r); // домножить матрицу модели на матрицу масштабирования в r раз
    
    (dragged?       // цвет меняется при перетаскивании:
      avo::ORANGE   // оранжевый для перетаскиваемого кружка,
    : avo::YELLOW)  // жёлтый для свободного
      .select(); 
    
    glCallList(LIST); // модель привязана к списку отрисовки
  }

  /// обработать движение
  void move(double time_step)
  {
    if (!dragged)
      p += time_step * v;
  }

  /// список отрисовки (display list), отвечающий кружку
  static GLuint LIST;
};

/// вывести текстовое представление кружка в поток вывода
std::ostream& operator<<(std::ostream &os, const Circle &c)
{
  return os << "position: "   << c.p
            << "\nvelocity: " << c.v
            << "\nmass: "     << c.m
            << "\nradius: "   << c.r  << std::endl;
}


//
// константы и глобальные переменные
//

/// количество кружков
const size_t N = 2;

/// коэффициент восстановления при ударе
const double k = 0.9;

/// ширина и высота окна в условных единицах и пикселях (для удобства захвата кружков мышью)
double width = 800.0, height = 600.0;
/// поправка координат мыши (координаты левого верхнего угла клиентской области окна в окне)
int mouse_origin_x, mouse_origin_y;
/// смещение захваченного кружка
int dragged_delta_x, dragged_delta_y;

/// сами кружки
Circle C[N];

/// индекс кружка, перетаскиваемого в данный момент мышью, либо -1 (если перетаскивание не происходит)
int dragged_mouse = -1;

/// таймер, используемый для замера временного шага между кадрами
Timer<> timer;

/// static-данные требуется явно размещать в некоторой единице трансляции
GLuint Circle::LIST;


//
// вспомогательные функции
//

/// следующее псевдослучайное double из отрезка [minv, maxv]
inline double random(double minv, double maxv)
{
  return minv + (maxv - minv) * (std::rand() / double(RAND_MAX));
}


/// создать псевдослучайные кружки, которые должны столкнуться в одной точке
void new_circles()
{
  std::cout << "\nGenerating new Circles:\n";

  // цель -- точка столкновения (cp от collision point)
  Vec2 cp(random(100.0, width - 100.0), random(100.0, height - 100.0));
  std::cout << "Collision point: " << cp << "\n";

  for (size_t i = 0; i < N; ++i)
  {
    Vec2 pos(random(0.0, width), random(0.0, height)),
         vel = cp - pos;
    
    std::cout << i << ")\n" <<
      (C[i] = Circle(pos, 0.25 * vel, // 4 секунды до столкновения
        /*масса*/random(0.1, 10.0), /*радиус*/random(4.0, 20.0)));
  }
}


//
// обработчики событий GLUT
//


/// перерисовка
void repaint()
{
  // очистить окно, закрасив его цветом очистки
  glClear(GL_COLOR_BUFFER_BIT);
  // нарисовать все кружки
  for (const auto &c : C)
    c.paint();

  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}


/// вычислить следующий кадр
void frame_move()
{
  double h = timer.step(); // время, прошедшее с предыдущего кадра
  if(h > 0.0)
  {
    // передвинуть все кружки
    for (auto &c : C)
      c.move(h);

    // проверить столкновение (в код зашито 2 кружка!)
    static_assert(N == 2, "fix the code below if you want more than 2 circles!");

    Circle &A = C[0], &B = C[1];
    if( (A.p - B.p).length() <= A.r + B.r ) // есть столкновение?
    {
      /////////////////////////////////////////////////////////////////////////
      // вычислить новые скорости
      // данные формулы не позволяют оперировать неподвижными объектами (с "бесконечной" массой)
      const double inv_mass  = 1.0 / (A.m + B.m);
      const Vec2   impulse   = A.m * A.v + B.m * B.v,
                   elastic_v = k * (A.v - B.v);

      A.v = inv_mass * (impulse - B.m * elastic_v);
      B.v = inv_mass * (impulse + A.m * elastic_v);

      // необходимо развести кружки, чтобы они перестали контактировать
            Vec2   AtoB = B.p - A.p;
      const double AtoBlength = AtoB.length();
      if (AtoBlength != 0.0)
      {
        const double dr = 1.00000001 * (A.r + B.r) - AtoBlength;
        if (dr > 0.0)
        {
          // разведём по вектору AtoB в пропорции, отвечающей импульсам тел
          const double impulseA = A.m * A.v.length(),
                       impulseB = B.m * B.v.length(),
                       alpha = impulseA / (impulseA + impulseB);

          AtoB *= dr / AtoBlength;
          // линейная интерполяция
          A.p -= (1.0 - alpha) * AtoB;
          B.p +=        alpha  * AtoB;
        }
      }
      else // центры шаров совпали -- патологическая ситуация
      {
        // развести как угодно, например, по 0x
        const double dx = 0.5 * (A.r + B.r);
        if (A.p.x < B.p.x)
          A.p.x -= dx, B.p.x += dx;
        else
          A.p.x += dx, B.p.x -= dx;
      }
    }
    
    // проверить, вылетели ли все кружки из области видимости
    {
      bool new_round = true;
      for (size_t i = 0; i < N; ++i)
      {
        const Circle &A = C[i];
        if (A.p.x > -A.r          &&
            A.p.y > -A.r          &&
            A.p.x < width  + A.r  &&
            A.p.y < height + A.r)
        {
          new_round = false;
          break;
        }
      }

      // если вылетели, сгенерировать новые шарики
      if(new_round)
        new_circles();
    }

    glutPostRedisplay(); // попросим GLUT выполнить перерисовку
  }
  
  // TASK: доработать код (в целом), для поддержки N > 2,
  // пусть '+' добавляет случайный круг 
  // (можно в пределах фиксированного максимума, т.е. N -- переменная "текущий размер",
  //  и есть константа, например, N_MAX, такая, что N < N_MAX),
  // а '-' удаляет некоторый круг (можно делать это просто как N--).
}


/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  mouse_origin_x = glutGet(GLUT_WINDOW_BORDER_WIDTH);  // ширина рамки окна
  mouse_origin_y = glutGet(GLUT_WINDOW_HEADER_HEIGHT); // высота заголовка окна -- не используется

  width  = w; // почему-то при повторном разворачивании-сворачивании
  height = h; // размеры окна постепенно уменьшаются...
  std::cout << "\nresized\n" << mouse_origin_x << ' ' << mouse_origin_y
            << '\n' << w << ' ' << h << '\n';

  // 2D "камера"
  glMatrixMode(GL_PROJECTION); // режим матрицы проекции
  glLoadIdentity();            // сбросить в единичную
  glOrtho(0.0, width, height, 0.0, -100.0, 100.0); // домножить на матрицу изометрической проекции,
  glMatrixMode(GL_MODELVIEW);  // режим матрицы модели

  // установить координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: /* ASCII-код ESC */ case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case ' ': case '\r': case '\n': // сгенерировать новый набор кружков
    new_circles();
    break;
  
  default:
    return;
  }
}


/** обработать клик мыши (нажата кнопка -- проверить, есть ли кружок под указателем, 
    отпущена -- освободить перетаскиваемый кружок) */
void mouse_click(int button, int state, int x, int y)
{
  if (button != GLUT_LEFT_BUTTON) // обрабатываем только левый клик
    return;

  // отпустить перетаскиваемый кружок
  if (dragged_mouse != -1)
  {
    C[dragged_mouse].dragged = false;
    dragged_mouse = -1;
  }

  // кнопка нажата, проверить, попал ли какой-нибудь кружок под указатель мыши
  if (state == GLUT_DOWN)
  {
    const Vec2 p(x, y); // координаты "мира" здесь совпадают с координатами в окне
    for (int i = 0; i < N; ++i)
    {
      Circle &A = C[i];
      const Vec2 delta = p - A.p;
      if (delta.length() <= A.r)
      {
        A.dragged = true;
        dragged_mouse = i;
        dragged_delta_x = int(delta.x);
        dragged_delta_y = int(delta.y);
        break;
      }
    }
  }
}


/// перемещение мыши, перетаскиваем кружок, (x, y) -- новые координаты указателя мыши
void mouse_move(int x, int y)
{
  if (dragged_mouse != -1)
    C[dragged_mouse].p = Vec2(x - dragged_delta_x, y - dragged_delta_y);
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE, сглаживание мультисемплингом MULTISAMPLE
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0);                    // координаты верхнего левого угла окна на экране
  glutInitWindowSize(int(width), int(height));     // полные ширина и высота окна
  glutCreateWindow("Circle collision simulation"); // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f); // установить цвет очистки (фон) -- чёрный
  avo::YELLOW.select(); // цвет материала по умолчанию (определено в Color.h)
  
  // изображение кружка (на самом деле полигональная аппроксимация сферы)
  Circle::LIST = glGenLists(1);        // получить идентификатор нового списка отрисовки
  glNewList(Circle::LIST, GL_COMPILE); // инициализировать этот список (режим "компиляции" команд), начать запись команд
  glutSolidSphere(1.0, 16, 4);         // выполнить рисование "кружка" (16 разбиений по "параллелям", 4 разбиения по "меридианам")
  glEndList();                         // закончить запись команд, скомпилировать список

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move);     // обработка движения
  glutDisplayFunc(repaint);     // перерисовка
  glutReshapeFunc(reshape);     // изменение размеров окна
  glutKeyboardFunc(handle_key); // обработка нажатий "обычных" клавиш
  glutMouseFunc(mouse_click);   // обработка нажатий кнопок мыши
  glutMotionFunc(mouse_move);   // обработка перемещения мыши (с зажатой кнопкой)
  // для отслеживания перемещения мыши с отпущенными кнопками можно зарегистрировать событие:
  // glutPassiveMotionFunc(mouse_move);

  // старт
  std::cout << "Starting Collision2D\n" << "k = " << k << "\n";
  new_circles();
  timer.start();  // инициализация таймера
  glutMainLoop(); // вход в цикл обработки событий
  std::cout << "\nbye!";
  return 0;
}
