﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Collision3D
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Симуляция движения и столкновения тяжёлых шаров в пространстве.
 * Шары притягиваются друг к другу под действием классического закона тяготения.
 * Нажатие SPACE/ENTER создаёт новые шары.
 * Нажатие ESC/Q -- выход из программы.
 * WASDZX и стрелки -- управление камерой.
 */
#include <GL/freeglut.h>
#include <cstdlib>        // rand/srand
#include <iostream>       

#include <avo/Timer.hpp>
#include <avo/Vec3.hpp>     // 3-мерный вектор
#include <avo/Color.hpp>    // RGBA-цвет в 128-битном формате (компоненты типа float)
#include <avo/SphericalCamera.hpp> // камера в сферических координатах ("турельного" типа)


using avo::Color;
using avo::Timer;
using avo::Vec3d;

//
// базовые определения
//

/* 
 * Вектор и камера вынесены в отдельные файлы, так что в этой секции размещён только Ball.
 * Замечу, что здешний Vec3 значительно "навороченнее", чем Vec2, так как взят из другого проекта.
 * В данном примере уровень использования функциональности Vec3 минимален.
 */

/// твёрдый тяжёлый шар
struct Ball
{
  Vec3d p; ///< текущая позиция
  Vec3d v; ///< текущая скорость
  Vec3d a_old; ///< старое ускорение (для метода Velocity Verlet)
  Vec3d a; ///< накопленное ускорение (сбрасывается при вычислении следующей позиции)
  double m;     ///< масса
  double r;     ///< радиус
  Color color;  ///< цвет (учитывается значение альфа-компонента -- в качестве коэффициента прозрачности)

  /// базовый конструктор (ускорение нулевое)
  explicit Ball
    (
      const Vec3d &pos  = Vec3d(),
      const Vec3d &vel  = Vec3d(),
      double       mass = 1.0, 
      double       rad  = 10.0,
      const Color &col  = avo::YELLOW
    )
      : p(pos), v(vel), m(mass), r(rad), color(col), a(), a_old()
  {}

  /// нарисовать шар
  void paint() const
  {
    glLoadIdentity();   // установить единичную матрицу в качестве матрицы модели
    glTranslated(p.x, p.y, p.z); // домножить матрицу модели на матрицу перемещения в (p.x, p.y, p.z)
    glScaled(r, r, r);  // домножить матрицу модели на матрицу масштабирования в r раз

    color.select(); // выставить нужный цвет
    glCallList(DISPLAY_LIST);  // вызвать список отрисовки (модель привязана к списку отрисовки)
  }

  /// вычислить позицию шара через time_step секунд
  void move(double time_step)
  {
    // используется метод Velocity Verlet: http://en.wikipedia.org/wiki/Velocity_Verlet#Velocity_Verlet 
    const auto half_step = 0.5 * time_step;
    const Vec3d ha = half_step * a;
    v += half_step * a_old + ha;
    p += time_step * (v + ha);
    a_old = a;
    a = Vec3d();
  }

  /// список отрисовки (display list), отвечающий шару
  static GLuint DISPLAY_LIST;
};

/// вывести текстовое представление шара в поток вывода
std::ostream& operator<<(std::ostream &os, const Ball &c)
{
  return os << "position: "    << c.p
            << "\nvelocity: "  << c.v
            << "\nmass: "      << c.m
            << "\nradius: "    << c.r   << std::endl;
}


//
// константы
// 

/// общее количество шаров
const size_t N = 100;

/// коэффициент восстановления при ударе
const double k = 0.8;

/// гравитационная постоянная
const double G = 4.0;

/// коэффициент "вязкого" трения (пропорционально квадрату радиуса и скорости)
const double F = 0.0;


//
// глобальные переменные
//

/// собственно гравитирующие тела (шары)
Ball C[N];

/// таймер, используемый для замера временного шага между кадрами
Timer<> timer;

/// счётчик кадров
int frame;

/// объект камеры
avo::SphericalCamera camera;

/// static-данные требуется явно размещать в некоторой единице трансляции
GLuint Ball::DISPLAY_LIST;


//
// вспомогательные функции
//

/// следующее псевдослучайное число из отрезка [minv, maxv]
template <class Float> // подходит и float и double, целые -- не подходят
inline Float random(Float minv, Float maxv)
{
  return minv + (maxv - minv) * (std::rand() / Float(RAND_MAX));
}


/// сгенерировать новый набор псевдослучайных шаров
void newBalls()
{
  std::cout << "\nGenerating new balls:\n";

  // две точки "столкновения" (в данный момент используются иначе)
  const Vec3d cp[2] = 
    {
      Vec3d(random(-500.0, 0.0), random(-200.0, 200.0), random(-200.0, 200.0))
    , Vec3d(random(0.0,  500.0), random(-200.0, 200.0), random(-200.0, 200.0))
    };

  std::cout << "Collision points: " << cp[0] << ", " << cp[1] << "\n";

  for (size_t i = 0; i < N; ++i)
  {
    Ball &b = C[i]; // заполним поля C[i]
    b.p = Vec3d(random(-600.0, 600.0), random(-600.0, 600.0), random(-600.0, 600.0)); // рандомная позиция
    b.v = -0.05 * (cp[i & 1] - b.p);  // одна половина летит к одной точке, другая -- к другой
    //b.v = Vec3();
    //b.v = Vec3((i & 1? -80: 80), 0, 0);
    b.r = random(4.0, 20.0); // radius
    b.m = random(2.0, 6.0) * b.r * b.r * b.r; // масса, случайная "плотность" вещества
    // случайный цвет
    b.color.red   = random(0.f, 1.f);
    b.color.green = random(0.f, 1.f);
    b.color.blue  = random(0.f, 1.f);
    // прозрачность
    b.color.alpha = random(0.3f, 1.f);

    if (N < 100)
      std::cout << i << ")\n" << b;
  }

  std::cout.flush();
}


/// вычислить и применить гравитационное притяжение для заданной пары шаров
inline void gravity(Ball &b1, Ball &b2)
{
  Vec3d r = b2.p - b1.p;
  double d = r.length();
  
  // притяжение работает только для непересекающихся шаров,
  // в противном случае применяется не зависящее от расстояния между центрами отталкивание
  // (увы, на практике это практически не помогает предотвратить вталкивание шаров друг в друга)
  const double r_sum = b1.r + b2.r;
  if (d < r_sum) d = -r_sum;

  r *= G / (d * d * d);
  b1.a += b2.m * r;
  b2.a -= b1.m * r;
}


/// применить "вязкое" трение
inline void friction(Ball &b)
{
  b.a -= F * b.r * b.r / b.m * b.v;
}


/// рассчитать столкновение двух шаров, возвращает признак наличия столкновения
void collide(Ball &a, Ball &b)
{
        Vec3d  dp = b.p - a.p;
  const double dplen = dp.length(),
               rsum = a.r + b.r;

  if (dplen <= rsum) // есть ли столкновение?
  {
    // следует ли пересчитывать скорости? (двигаются ли шары навстречу друг другу)
    if (0.0 < avo::dot(dp, a.v - b.v)) // dot -- скалярное произведение
    {
      const double inv_mass = 1.0 / (a.m + b.m);
      const Vec3d impulse = a.m * a.v + b.m * b.v;
      const Vec3d elastic_v = k * (a.v - b.v);
      a.v = inv_mass * (impulse - b.m * elastic_v);
      b.v = inv_mass * (impulse + a.m * elastic_v);
    }

    // необходимо развести кружки, чтобы они перестали контактировать
    if (dplen != 0.0)
    {
      const double dr = 1.00000001 * rsum - dplen;
      if (dr > 0.0)
      {
        // разведём по вектору AtoB в пропорции, отвечающей импульсам тел
        const double impulseA = a.m * a.v.length(),
                     impulseB = b.m * b.v.length(),
                     alpha = impulseA / (impulseA + impulseB);

        dp *= dr / dplen;
        // линейная интерполяция
        a.p -= (1.0 - alpha) * dp;
        b.p +=        alpha  * dp;
      }
    }
    else // центры шаров совпали -- патологическая ситуация
    {
      // развести как угодно, например, по 0x
      const double dx = 0.5 * rsum;
      if (a.p.x < b.p.x)
        a.p.x -= dx, b.p.x += dx;
      else
        a.p.x += dx, b.p.x -= dx;
    }
  }
}


//
// обработчики событий GLUT
//

/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  // новая матрица проекции
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();    // удалить предыдущую матрицу из стека матриц (стек матриц используется SphericalCamera)
  glLoadIdentity(); // загрузить единичную матрицу
  /* угол зрения по вертикали, соотношение сторон экрана, минимальная дальность, максимальная дальность */
  gluPerspective(75, double(w) / double(h), 16.0, 4096.0); // домножить на матрицу перспективной коррекции
  glPushMatrix();   // затолкнуть текущую матрицу в стек матриц (копируется, текущая остаётся на вершине)
  camera.apply();   // применить к текущей матрице матрицу камеры

  // установить координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}

/// перерисовка
void repaint()
{
  // очистить окно, закрасив его цветом очистки, сбросить буфер глубины
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // нарисовать все шары
  for (const auto &c : C)
    c.paint();

  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}


/// вычислить состояние сцены на следующий кадр
void frame_move()
{
  const double MAX_STEP = 0.05;
  const int INTEGRATOR_ROUNDS = 4;

  double h = timer.step(); // время, прошедшее с предыдущего кадра
  if (h <= 0.0) return;
  // будем "тормозить", если компьютер не справляется, ограничив сверху шаг по времени
  if (h > MAX_STEP) h = MAX_STEP;

  for (int i = 0; i < N; ++i) // "треугольный" цикл
  {
    Ball &Ci = C[i];
    for (size_t j = i + 1; j < N; ++j)
      collide(Ci, C[j]); // проверить и рассчитать столкновение
  }

  // в ti накапливается суммарное количество движения (total impulse -- "полный импульс")
  Vec3d ti;
  // вычисляем и выводим полный импульс каждый 512-й кадр
  if ((frame++ & 511) == 0)
  {
    for (const auto &c : C)
      ti += c.m * c.v;
    std::cout << "total impulse = " << ti << std::endl;
  }

  // делаем дополнительные шаги интегрирования для данного временного шага (это делать необязательно)
  h /= INTEGRATOR_ROUNDS;

  // применить силы, 
  // здесь бы могли пригодиться какие-нибудь продвинутые алгоритмы
  // либо для ускорения работы, либо для улучшения качества (например, компенсационное суммирование)
  for (int round = 0; round < INTEGRATOR_ROUNDS; ++round)
  {
    // вычисляем новые ускорения шаров
    for (int i = 0; i < N; ++i)
    {
      Ball &Ci = C[i];
      for (size_t j = i + 1; j < N; ++j)
        gravity(Ci, C[j]);
      friction(Ci);
    }

    // интегрируем ускорения, получая новые позиции
    for (auto &c : C)
    {
      c.move(h);
      if (c.p.dotSquare() > 1e+9) // если шар улетел слишком далеко от начала координат,
      {                              // то возвращаем его в окрестность начала координат
        c.p = Vec3d(random(-300.0, 300.0), random(-300.0, 300.0), random(-300.0, 300.0));
        c.v = Vec3d();
      }
    }
  }

  // уведомим GLUT о необходимости перерисовать сцену
  glutPostRedisplay();
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void key_press(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 'w': case 'W': case 'ц': case 'Ц': // ближе к началу координат
    camera.forward();
    break;

  case 's': case 'S': case 'ы': case 'Ы': // дальше от начала координат
    camera.backward();
    break;

  case 'a': case 'A': case 'ф': case 'Ф': // переместить влево
    camera.left();
    break;

  case 'd': case 'D': case 'в': case 'В': // переместить вправо
    camera.right();
    break;

  case 'z': case 'Z': case 'я': case 'Я':
    camera.up();
    break;

  case 'x': case 'X': case 'ч': case 'Ч':
    camera.down();
    break;

  case 27: case 'q': case 'Q': case 'й': case 'Й':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    return; // и покинуть данную функцию...

  case ' ': case '\r': case '\n':
    newBalls(); // сделать новые шарики
    break;
  
  default:
    return; // остальные игнорируются
  }

  camera.apply(); // применить матрицу камеры к матрице проекции
  glutPostRedisplay();
}


/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void special_key_press(int key, int x, int y)
{
  // движение камеры аналогично WASD, только с помощью стрелок на клавиатуре
  switch (key)
  {
  case GLUT_KEY_LEFT:
    camera.left();
    break;

  case GLUT_KEY_UP:
    camera.up();
    break;

  case GLUT_KEY_RIGHT:
    camera.right();
    break;

  case GLUT_KEY_DOWN:
    camera.down();
    break;

  default:
    return; // остальные игнорируются
  }

  camera.apply(); // применить матрицу камеры к матрице проекции
  glutPostRedisplay();
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  // режим:
  glutInitDisplayMode(GLUT_RGBA        // 32-битный цвет RGBA,
                    | GLUT_ALPHA       // поддержка альфа-канала для имитации прозрачности
                    | GLUT_DEPTH       // поддержка буфера глубины
                    | GLUT_MULTISAMPLE // сглаживание мультисемплингом (FS MSAA)
                    | GLUT_DOUBLE);    // двойная буферизация
  glutInitWindowPosition(0, 0);   // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600);   // полные ширина и высота окна
  glutCreateWindow("Ball collision simulation"); // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glClearColor(0.0, 0.0, 0.0, 1.0); // установить цвет очистки (фон) -- чёрный, непрозрачный
  glEnable(GL_COLOR_MATERIAL);      // выбирать цвет материала командой glColor
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты

    glEnable(GL_BLEND); // включить интерпретацию альфа канала как коэффициента смешивания цветов (для имитации прозрачности)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // выбрать формулу смешивания
    /* sfactor == GL_SRC_ALPHA -- домножить цвет нового пикселя на его альфу, 
       dfactor == GL_ONE_MINUS_SRC_ALPHA -- домножить цвет старого пикселя на (1 - альфа нового),
       результирующий цвет получается сложением этих двух цветов */

  glEnable(GL_CULL_FACE);  // включить отбрасывание полигонов, смотрящих "от" камеры (повёрнутых обратной стороной)
  glEnable(GL_DEPTH_TEST); // использовать буфер глубины по назначению
  glEnable(GL_NORMALIZE);  // форсировать нормирование нормалей (при масштабировании сфер происходит масштабирование и нормалей)
  glEnable(GL_LIGHTING);   // включить повершинное освещение
  glEnable(GL_LIGHT0);     // включить источник света №0

  // сгенерировать модель шара (полигональная аппроксимация сферы)
  Ball::DISPLAY_LIST = glGenLists(1);        // получить идентификатор нового списка отрисовки
  glNewList(Ball::DISPLAY_LIST, GL_COMPILE); // инициализировать этот список (режим "компиляции" команд), начать запись команд
  glutSolidSphere(1.0, 16, 8);       // выполнить рисование "кружка" (16 разбиений по "параллелям", 8 разбиения по "меридианам")
  glEndList();                       // закончить запись команд, скомпилировать список

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move);            // обработка движения
  glutDisplayFunc(repaint);            // перерисовка
  glutReshapeFunc(reshape);            // изменение размеров окна
  glutKeyboardFunc(key_press);         // обработка нажатий "обычных" клавиш
  glutSpecialFunc(special_key_press);  // обработка нажатий "специальных" клавиш

  // старт
  std::cout << "Starting Collision...\n" << "k = " << k << "\n";
  newBalls();
  timer.start();  // инициализация таймера
  glutMainLoop(); // вход в цикл обработки событий
  std::cout << "\nbye!";
  return 0;
}
