﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  Contour.cpp
 * @brief реализация Contour.h
 */
#include "Contour.hpp"
#include <algorithm>
#include <iterator>
#include <numeric>
#include <functional>


// вычислить ориентированную площадь
Scalar Contour::area() const
{
  if (size() > 2u)
  {
    return Scalar(0.5) *
      std::inner_product
        (
          begin(),
          end() - 1,
          begin() + 1,
          cross(back(), front()),
          std::plus<Scalar>(),
          avo::cross<Scalar>
        );
  }
  return 0.0;

  /*
  Этот код выполняет то же, что и нижеприведённый

      const_iterator b = begin(), a = b++;
      double S = cross(back(), front()); // замыкающий
    
      // сумма площадей ориентированных треугольников
      for (const const_iterator b_end = end(); b != b_end; a = b++)
        S += cross(*a, *b);
        
      return 0.5 * S;

  Смысл данной формулы следующий.

  Пусть есть треугольник ABC. 
  Его площадь равна половине произведения основания на высоту.
  Например, AB -- основание, высоту мы можем получить, умножив AC на sin <CAB.
  Или иначе S = 1/2 ||B - A|| ||C - A|| sin <CAB.
  (Эта площадь берётся со знаком -- "ориентированная площадь".)

  Утверждение. Пусть есть 2D вектора a и b.
  Тогда ||a|| ||b|| sin <ab = ax by - ay bx = a * b ("псевдоскалярное произведение").
  (Угол отсчитывается поворотом от a к b. Положительное направление -- против ЧС.)
  Данная операция антикоммутативна: a * b = -b * a.
    
  Пусть есть выпуклый многоугольник OA1A2A3...An.
  Очевидно, его площадь есть сумма площадей треугольников OA1A2, OA2A3, ..., OAn-1An
  или (полагая, что обход Ai идёт против ЧС):
  (1)  S = 1/2 sum { (Ai-1 - O) * (Ai - O) | i=2..n }.

  Теперь предположим, что O -- не вершина многоугольника, а просто произвольная
  точка на плоскости, например, начало используемой системы координат (т.е. положим A - O = A).
  Тогда из правой части (1) надо вычесть площадь треугольника OA1An, а значит
  площадь многоугольника A1A2...An будет равна
  (2)  S = 1/2 (sum { Ai-1 * Ai | i=2..n } + An * A1). // контур "замкнулся"

  Утверждение. Формула (2) верна и для невыпуклого многоугольника, заданного
  контуром A1A2...An, при обходе которого в заданном порядке внутренность многоугольника
  всегда находится слева.

  Утверждение. Пусть Ai = (xi, yi), i=1..n.
  (3)  S = 1/2 (sum { (xi + xi-1)(yi - yi-1) | i=2..n } + (x1 + xn)(y1 - yn))
    
  Распишем сумму.
  sum { (xi + xi-1)(yi - yi-1) | i=2..n } =
  sum { (xi-1 yi - xi yi-1) + (xi yi) - (xi-1 yi-1) | i=2..n } =
  sum { Ai-1 * Ai | i=2..n } + sum { xi yi | i=2..n } - sum { xi yi | i=1..n-1 } =
  sum { Ai-1 * Ai | i=2..n } + (xn yn - x1 y1),
  что отличается от (2) на
  (x1 y1 - xn yn) + An * A1 = x1 y1 - xn yn + xn y1 - x1 yn = (x1 + xn)(y1 - yn).

  Конструктивное обоснование (3) (площади трапеций) см., например, здесь:
  http://www.mathopenref.com/coordpolygonarea2.html

  Что лучше: (2) или (3)?
  В те времена, когда умножение выполнялось значительно медленнее, чем сложение,
  формула (3) могла дать преимущество в скорости вычисления относительно формулы (2):
  (3n-1) сложение и n умножений vs (2n-1) сложения и 2n умножений.

  С точки зрения накопления погрешности всё не так очевидно.
  Могут быть как серьёзные колебания в величинах x и y, приводящие к большой
  относительной погрешности слагаемых в (3), так и серьёзные колебания в площади
  треугольников в (2) (какие-то треугольники могут быть вырожденными).

  Рассмотрим погрешность метода: предположим, что все вычисления точны, но
  данные переданы с погрешностью.
  Пусть все координаты x', y' отличаются от истинных x, y по модулю не более, чем на z (малое число).
  (2):
  (a + u) (b + v) - a b = u b + v a + u v, за малостью можно пренебречь u v, <= z (|a| + |b|)
  (x'i y'i-1 - x'i-1 y'i) - (xi yi-1 - xi-1 yi) =
  (x'i y'i-1 - xi yi-1) - (x'i-1 y'i - xi-1 yi) <=
    z (|xi| + |xi-1| + |yi| + |yi-1|),

  что даёт оценку погрешности
  z sum { |xi| + |yi| | i=1..n }. // 1/2 уничтожилась благодаря повторению |xi-1| и |yi-1| под суммой

  (3):
  (x'i + x'i-1)(y'i - y'i-1) - (xi + xi-1)(yi - yi-1) = // xi + xi-1 = a, yi - yi-1 = b
  (a + 2u)(b + 2v) - a b <= 2z (|a| + |b|),

  что даёт оценку погрешности
  z sum { |xi + xi-1| + |yi - yi-1| | i=2..n } + ..., что, увы, можно мажорировать
  2z sum { |xi| + |yi| | i=1..n },

  т.е. вдвое больше, чем для (2). Впрочем, проведённый анализ весьма поверхностен.
  Возможно, стоило бы произвести замеры на реальных примерах...     
  */
}


// вычислить периметр
Scalar Contour::perimeter() const
{
  return std::inner_product
    (
      begin() + 1,
      end(),
      begin(),
      distance(back(), front()),
      std::plus<Scalar>(),
      avo::distance<Scalar>
    );
}


std::ostream& operator<<(std::ostream &os, const Contour &c)
{
  os << '\n' << c.size() << '\n';
  // почему-то студия не хочет компилировать std::ostream_iterator<Vec2>
  // std::copy (c.begin(), c.end(), std::ostream_iterator<Vec2>(os, "\n"));
  for (const auto &p : c)
    os << p << '\n';
  return os;
}


std::istream& operator>>(std::istream &is, Contour &c)
{
  c.clear();
  if (is.good())
  {
    // прочитать количество вершин
    Contour::size_type n = 0, i = 0;
    is >> n;
    // зарезервировать место -- может бросить исключение
    c.resize (n);
    
    for (; i != n && is.good(); ++i)
      is >> c[i];
    // установить реальный размер
    c.resize (i);
  }

  return is;
}
