﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  Contour.hpp
 * @brief ориентированный двумерный контур (многоугольник)
 */
#pragma once

#include <vector>
#include <memory>
#include <istream>
#include <ostream>

#include <avo/Vec2.hpp>
#include <avo/Bbox2.hpp>


/// тип скаляра
typedef double Scalar;
/// тип вектора, используемый в GLUPoly
typedef avo::Vec2<Scalar> Vec2;
/// прямоугольник со сторонами, параллельными осям координат
typedef avo::Bbox2<Scalar> Bbox2;

/// ориентированный двумерный контур, основан на std::vector
/**
 * Расширение используется для того, чтобы "примешать" к вектору дополнительные операции,
 * но так как std::vector не предназначен служить базой наследования, используется private-расширение.
 */
class Contour : private std::vector<Vec2>
{ 
  typedef std::vector<Vec2> Base;

public:

  // то, что ожидает увидеть пользователь std::vector
  using Base::iterator;
  using Base::const_iterator;
  using Base::reverse_iterator;
  using Base::const_reverse_iterator;
  
  using Base::allocator_type;
  using Base::value_type;
  using Base::reference;
  using Base::const_reference;
  using Base::pointer;
  using Base::const_pointer;
  using Base::difference_type;
  using Base::size_type;

  using Base::begin;
  using Base::end;
  using Base::cbegin;
  using Base::cend;
  using Base::rbegin;
  using Base::rend;
  using Base::crbegin;
  using Base::crend;

  using Base::size;
  using Base::max_size;
  using Base::resize;
  using Base::capacity;
  using Base::empty;
  using Base::reserve;
  using Base::data;
  using Base::shrink_to_fit;
    
  using Base::operator[];
  using Base::at;
  using Base::front;
  using Base::back;

  using Base::assign;
  using Base::push_back;
  using Base::emplace_back;
  using Base::pop_back;
  using Base::insert;
  using Base::emplace;
  using Base::erase;
  using Base::clear;
  using Base::get_allocator;

  // конструкторы и присваивание

  Contour() {}
  explicit Contour(size_type n)
    : Base(n) {}

  Contour(size_type n, const value_type &val)
    : Base(n, val) {}

  Contour(const Contour &other)
    : Base(other) {}

  Contour(Contour &&other)
    : Base(std::move(static_cast<Base&>(other))) {}

  template <class InIt>
  Contour(InIt from, InIt to)
    : Base(from, to) {}

  Contour& operator=(const Contour &other)
  {
    Base::operator=(other);
    return *this;
  }

  Contour& operator=(Contour &&other)
  {
    Base::operator=(std::move(static_cast<Base&>(other)));
    return *this;
  }

  // прочие операции

  /// обмен
  void swap(Contour &other)
  {
    Base::swap(static_cast<Base&>(other));
  }

  /// вычислить ориентированную площадь
  Scalar area() const;

  /// вычислить периметр
  Scalar perimeter() const;

  /// generate axes-aligned bounding box (O(N) operation)
  Bbox2 box() const
  {
    return Bbox2(begin(), end());
  }

  // сравнения

  bool operator==(const Contour &other) const
  {
    return *(const Base*)this == (const Base&)other;
  }

  bool operator!=(const Contour &other) const
  {
    return !(*this == other);
  }

  /// параллельный сдвиг
  Contour& operator+=(const Vec2 &offset)
  {
    Vec2 *data = &front();
    std::size_t sz = size();
    for (std::size_t i = 0; i < sz; ++i)
      data[i] += offset;
    return *this;
  }

  /// параллельный сдвиг
  Contour& operator-=(const Vec2 &offset)
  {
    Vec2 *data = &front();
    std::size_t sz = size();
    for (std::size_t i = 0; i < sz; ++i)
      data[i] -= offset;
    return *this;
  }

  /// изотропное масштабирования
  Contour& operator*=(Scalar scale)
  {
    Vec2 *data = &front();
    std::size_t sz = size();
    for (std::size_t i = 0; i < sz; ++i)
      data[i] *= scale;
    return *this;
  }

  /// анизотропное масштабирование
  Contour& operator*=(const Vec2 &scale_xy)
  {
    Vec2 *data = &front();
    std::size_t sz = size();
    for (std::size_t i = 0; i < sz; ++i)
      data[i] *= scale_xy;
    return *this;
  }
};


/// swap function
inline void swap(Contour &a, Contour &b)
{
  a.swap(b);
}


inline std::ostream& operator<<(std::ostream &os, const Vec2 &p)
{
  return os << p.x << ' ' << p.y;
}

inline std::istream& operator>>(std::istream &is, Vec2 &p)
{
  return is >> p.x >> p.y;
}


std::ostream& operator<<(std::ostream &os, const Contour&);
std::istream& operator>>(std::istream &is, Contour&);
