﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  Mesh2.cpp
 * @brief реализация Mesh.h, опирается на тесселятор GLU
 */
#include <list>
#include <algorithm>
#include <memory>

// включает GLU через FreeGlut
#include "Mesh2.hpp"


//
// колбеки тесселятора GLU

#ifndef CALLBACK // WinAPI подставляет __stdcall
# define CALLBACK
#endif

typedef void (CALLBACK *tessCallback)();
void CALLBACK beginData(GLenum mode, void *data);
void CALLBACK endData(void *data);
void CALLBACK vertexData(void *vertex, void *data);
void CALLBACK combineData
  (
    GLdouble coords[3], // здесь нет выбора в возможных типах данных
    void    *vertex[4],
    GLfloat  weight[4],
    void   **outData,
    void    *data
  );


/// обёртка GLUtesselator
class Tessellator
{
  friend class Mesh2;
  GLUtesselator *tess;

  std::list<Prim2> prims; ///< набранные примитивы
  std::list<Vec2>  ip;    ///< найденные точки пересечения контуров

public:

  /// создаёт тесселятор GLU и привязывает колбеки, правило позволяет выбрать геометрическую операцию
  explicit Tessellator(GLenum rule = GLU_TESS_WINDING_ODD)
    : tess(gluNewTess())
  {
    if (tess)
    {
      gluTessProperty (tess, GLU_TESS_WINDING_RULE, rule);
      gluTessCallback (tess, GLU_TESS_BEGIN_DATA  , (tessCallback) beginData);
      gluTessCallback (tess, GLU_TESS_END_DATA    , (tessCallback) endData);
      gluTessCallback (tess, GLU_TESS_VERTEX_DATA , (tessCallback) vertexData);
      gluTessCallback (tess, GLU_TESS_COMBINE_DATA, (tessCallback) combineData);
    }
  }

  /// удаляет тесселятор GLU
  ~Tessellator()
  {
    gluDeleteTess (tess);
  }

  /// начать накопление нового примитива
  void begin(GLenum mode)
  {
    prims.emplace_back (mode);
  }

  /// добавить вершину к примитиву
  void vertex(const Vec2 &v)
  {
    prims.back().vertices().push_back (v);
  }

  /// создать новую точку пересечения
  Vec2& allocip()
  {
    ip.push_back (Vec2());
    return ip.back();
  }
};

/// GLU tesselator begin callback (начать примитив)
void CALLBACK beginData(GLenum mode, void *data)
{
  static_cast<Tessellator*>(data)->begin (mode);
}

/// GLU tesselator end callback (закончить примитив)
void CALLBACK endData(void *data)
{
  // does nothing
}

/// GLU tesselator (add a) vertex callback (добавить вершину)
void CALLBACK vertexData(void *vertex, void *data)
{
  static_cast<Tessellator*>(data)->vertex (*(const Vec2*)vertex);
}

/// GLU tesselator combine callback (вызывается в ситуации обнаружения точки пересечения)
void CALLBACK combineData
  (
    GLdouble coords[3],
    void    *vertex[4],
    GLfloat  weight[4],
    void   **outData,
    void    *data
  )
{
  // просто копирует точку пересечения, т.к. у нас нет никаких данных,
  // которые нужно было бы, например, интерполировать в новой вершине
  Vec2 &p = static_cast<Tessellator*>(data)->allocip();
  p.x = coords[0];
  p.y = coords[1];
  *outData = &p;
}


// уничтожить тесселятор GLU
void Mesh2::releaseTessellator()
{
  delete tess;
  tess = nullptr;
}


// создать Tessellator и начать многоугольник
void Mesh2::startPolygon()
{
  if (!tess)
    releaseTessellator();

  tess = new Tessellator(); // возможные ошибки игнорируются
  gluTessBeginPolygon (tess->tess, tess);
}


// собрать накопленные примитивы, уничтожить Tessellator
void Mesh2::endPolygon()
{
  gluTessEndPolygon (tess->tess);

  // собрать примитивы
  for (auto &p : tess->prims)
    body.emplace_back (std::move(p));

  releaseTessellator();
}


void Mesh2::startContour()
{
  gluTessBeginContour (tess->tess);
}

void Mesh2::endContour()
{
  gluTessEndContour (tess->tess);
}

void Mesh2::pushPoint (const Vec2 &p)
{
  double point[] = { p.x, p.y, 0.0, 0.0 };
  gluTessVertex (tess->tess, point, const_cast<Vec2*>(&p));
}


void Mesh2::swap (Mesh2 &other)
{
  using std::swap;
  swap (tess, other.tess);
  swap (externalOutline, other.externalOutline);
  swap (internalOutline, other.internalOutline);
  swap (background, other.background);
  excont.swap (other.excont);
  incont.swap (other.incont);
  body.swap   (other.body);
}
