﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  Mesh2.hpp
 * @brief полигональный примитив для OpenGL
 */
#pragma once

#include <type_traits>
#include <vector>
#include <avo/Color.hpp>
#include "Contour.hpp"

// включает gl.h через FreeGlut
#include <GL/freeglut.h>


/// 2D-примитив
class Prim2
{
public:

  typedef std::vector<Vec2> Vertices;

  /// по умолчанию -- просто набор точек
  explicit Prim2(GLenum mode = GL_POINTS)
    : mode(mode) {}

  /// инициализировать из последовательности, режим по умолчанию -- набор точек
  template <class ForIt>
  Prim2(ForIt first, ForIt last, GLenum mode = GL_POINTS)
    : mode(mode), data(first, last) {}

  // compiler should define the following ctors and op=s automatically, but MSVC2010 doesn't do it

  Prim2(const Prim2 &other)
    : mode(other.mode), data(other.data) {}
  Prim2(Prim2 &&other)
    : mode(other.mode), data(std::move(other.data)) {}

  Prim2& operator=(const Prim2 &other)
  {
    mode = other.mode;
    data = other.data;
    return *this;
  }

  Prim2& operator=(Prim2 &&other)
  {
    mode = other.mode;
    data = std::move(other.data);
    return *this;
  }

  /// обменять содержимое
  void swap(Prim2 &other)
  {
    using std::swap;
    swap (mode, other.mode);
    data.swap (other.data);
  }

  /// получить доступ к вершинам примитива
  Vertices& vertices() { return data; }
  const Vertices& vertices() const { return data; }

  /// отрисовка примитива, использует массив вершин (OpenGL 1.1)
  void paint() const
  {
    static_assert (std::is_same<Scalar, float>::value || std::is_same<Scalar, double>::value,
      "GLUPoly: only double and float are supported as a Scalar type.");
    glVertexPointer (2, (sizeof(Scalar) == sizeof(float)? GL_FLOAT: GL_DOUBLE), sizeof(Vec2), &(data[0]));
    glDrawArrays (mode, 0, data.size());
  }

private:

  /// режим отрисовки примитива OpenGL (тот же смысл, что у параметра glBegin)
  GLenum mode;
  /// вершины, составляющие примитив (только пары координат)
  Vertices data;
  friend class Mesh2;
};


inline void swap(Prim2 &a, Prim2 &b)
{
  a.swap (b);
}


/// набор примитивов, задающий произвольный плоский многоугольник
class Mesh2
{
  std::vector<Prim2>
    excont, ///< внешние контуры (каждый ограничивает компоненту связности)
    incont, ///< внутренние контуры ("дырки")
    body    ///< треугольники, на которые разбита внутренность многоугольника (триангуляция)
  ;

  /// выставляет цвета по умолчанию
  void setDefaultColors()
  {
    externalOutline = avo::YELLOW;
    internalOutline = avo::ORANGE;
    background = avo::GREEN;
  }

public:

  avo::Color
    externalOutline, ///< цвет внешних контуров
    internalOutline, ///< цвет внутренних контуров
    background       ///< цвет заливки внутренности многоугольника
  ;


  /// конструктор по умолчанию
  Mesh2()
    : tess(nullptr)
  {
    setDefaultColors ();
  }

  /// построить многоугольник на основе набора контуров
  /**
   * Выполняет триангуляцию и сортировку контуров (на внешние и внутренние).
   */
  template <class ContourForIt>
  Mesh2(ContourForIt contbegin, ContourForIt contend)
    : tess(nullptr)
  {
    setDefaultColors ();
    if (contbegin != contend)
    {
      startPolygon ();
      // добавить каждый контур в триангуляцию
      for (; contbegin != contend; ++contbegin)
      {
        auto pbegin = contbegin->begin(),
             pend = contbegin->end();

        if (pbegin != pend) // контур не пуст
        {
          // определить ориентацию контура по знаку площади
          ( contbegin->area() < 0.0? incont: excont )
            .emplace_back (Prim2(pbegin, pend, GL_LINE_LOOP));

          // триангуляция
          startContour ();
          for (; pbegin != pend; ++pbegin) // передать все вершины контура
            pushPoint (*pbegin);
          endContour ();
        }
      }
      endPolygon ();
    }
  }

  // compiler might have defined the following ctors and op=s automatically if a destructor wasn't there

  Mesh2(const Mesh2 &other)
    : excont(other.excont),
      incont(other.incont),
      body(other.body),
      externalOutline(other.externalOutline),
      internalOutline(other.internalOutline),
      background(other.background),
      tess(nullptr)
  {}

  Mesh2(Mesh2 &&other)
    : excont(std::move(other.excont)),
      incont(std::move(other.incont)),
      body(std::move(other.body)),
      externalOutline(other.externalOutline),
      internalOutline(other.internalOutline),
      background(other.background),
      tess(nullptr)
  {}

  Mesh2& operator=(const Mesh2 &other)
  {
    excont = other.excont;
    incont = other.incont;
    body = other.body;
    externalOutline = other.externalOutline;
    internalOutline = other.internalOutline;
    background = other.background;
    return *this;
  }

  Mesh2& operator=(Mesh2 &&other)
  {
    excont = std::move(other.excont);
    incont = std::move(other.incont);
    body = std::move(other.body);
    externalOutline = other.externalOutline;
    internalOutline = other.internalOutline;
    background = other.background;
    return *this;
  }


  // наличие явно определённого деструктора -- банальная перестраховка
  ~Mesh2()
  {
    releaseTessellator (); // избыточно...
  }


  /// обмен содержимым
  void swap(Mesh2 &other);

  /// отрисовка средствами OpenGL
  void paint() const
  {
    blockPaint (body.begin(), body.end(), background);
    blockPaint (incont.begin(), incont.end(), internalOutline);
    blockPaint (excont.begin(), excont.end(), externalOutline);
  }

private:

  /// нарисовать набор примитивов общим цветом
  template <class ForIt>
  static void blockPaint(ForIt begin, ForIt end, const avo::Color &color)
  {
    color.select ();
    for (; begin != end; ++begin)
      begin->paint ();
  }

  // используется только в конструкторе из контуров, в остальное время == nullptr
  class Tessellator *tess;

  void releaseTessellator();
  void startPolygon(); ///< выполняет также привязку колбеков GLU
  void endPolygon();   ///< выполняет отвязку колбеков GLU
  void startContour();
  void endContour();
  void pushPoint(const Vec2&);
};


inline void swap(Mesh2 &a, Mesh2 &b)
{
  a.swap (b);
}
