/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file   Bbox3.hpp
 * @brief  3D axes-aligned bounding box
 * @author Dmitri Kuvshinov
 *
 * Converted from Bbox2.hpp
 */
#pragma once
#include <avo/Vec3.hpp>


namespace avo
{

/// 3D axes-aligned bounding box
template <class Float>
struct Bbox3
{
  typedef Bbox3<Float> This;
  typedef Vec3<Float>  Vec;

  Vec pmin ///< left-lower point
    , pmax ///< right-upper point
    ;

  /// origin
  Bbox3() {}
  
  /// a point
  explicit Bbox3(const Vec &p)
    : pmin (p), pmax (p) {}

  /// minimal bbox containing both a and b
  Bbox3(const Vec &a, const Vec &b)
    : pmin (a), pmax (b)
  {
    pmax.maximize(a);
    pmin.minimize(b);
  }

  /// minimal bbox containing all points from the memory range
  /**
   * Should come out an optimized version against InputIterator-based template analogue below.
   */
  Bbox3(const Vec *begin, const Vec *end)
    : pmin (Vec()), pmax (Vec())
  {
    bbox3(begin, end - begin, pmin, pmax);
  }

  /// minimal bbox containing all points from the range
  template <class InIt>
  Bbox3(InIt begin, InIt end)
  {
    if (begin != end)
    {
      pmax = pmin = *begin;
      while (++begin != end)
        grow(*begin);
    }
  }


  /// checks if the box has no area ("is empty")
  bool empty() const
  {
    return pmin == pmax;
  }

  /// checks if the given point falls inside the box
  bool contains(const Vec &p) const
  {
    return p.dominates(pmin) && pmax.dominates(p);
  }

  /// checks if the given box is inside this box
  bool contains(const Bbox3 &b) const
  {
    return contains(b.pmin) && contains(b.pmax);
  }

  /// width of the box (along 0x axis)
  Float width() const
  {
    return pmax.x - pmin.x;
  }

  /// height of the box (along 0y axis)
  Float height() const
  {
    return pmax.y - pmin.y;
  }

  /// depth of the box (along 0z axis)
  Float depth() const
  {
    return pmax.z - pmin.z;
  }

  /// calculate width and height simultaneously
  Vec extents() const
  {
    return pmax - pmin;
  }

  /// move the box by the given delta vector
  This& move(const Vec &delta)
  {
    pmin += delta;
    pmax += delta;
    return *this;
  }

  /// grow the box to contain the given point
  This& grow(const Vec &p)
  {
    pmin.minimize(p);
    pmax.maximize(p);
    return *this;
  }

  /// grow with another bbox3 ("join")
  This& grow(const This &b)
  {
    pmin.minimize(b.pmin);
    pmax.maximize(b.pmax);
    return *this;
  }

  /// intersect with another bbox3
  This& intersect(const This &b)
  {
    pmin.maximize(b.pmin);
    pmax.minimize(b.pmax);
    return *this;
  }

  /// calculates biggest diagonal of the box
  Float diagonal() const
  {
    return (pmin - pmax).length();
  }

  /// calculates volume of the box
  Float volume() const
  {
    const Vec ext = extents();
    return ext.x * ext.y * ext.z;
  }

  /// calculates area of box surface
  Float surfaceArea() const
  {
    const Vec ext = extents();
    return Float(2.0) * (ext.x * ext.y + ext.y * ext.z + ext.x * ext.z);
  }

  /// box of the union of this and other boxes
  This& operator+=(const This &other)
  {
    return grow(other);
  }

  /// intersection of two boxes
  This& operator&=(const This &other)
  {
    return intersect(other);
  }
};


template <class Float>
inline bool operator==(const Bbox3<Float> &a, const Bbox3<Float> &b)
{
  return a.pmin == b.pmin && a.pmax == b.pmax;
}

template <class Float>
inline bool operator!=(const Bbox3<Float> &a, const Bbox3<Float> &b)
{
  return !(a == b);
}

template <class Float>
inline std::ostream& operator<<(std::ostream &os, const Bbox3<Float> &box)
{
  return os << box.pmin << " -- " << box.pmax;
}


/// build box of the union of two boxes
template <class Float>
inline Bbox3<Float> operator+(const Bbox3<Float> &a, const Bbox3<Float> &b)
{
  return (Bbox3<Float>(a)) += b;
}

/// build an intersection of two boxes (box again)
template <class Float>
inline Bbox3<Float> operator&(const Bbox3<Float> &a, const Bbox3<Float> &b)
{
  return (Bbox3<Float>(a)) &= b;
}

}
