/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file SphericalCamera.hpp
 * @author Dmitri Kuvshinov
 * @brief spherical coordinates to simulate rotating camera
 *
 */
#pragma once
#include <GL/freeglut.h> // for OpenGL functions
#include <cmath>


namespace avo
{

const double PI       = 3.14159265358979323846;
const double PI_div_2 = 1.57079632679489661923;
const double PI_mul_2 = 2.0 * PI;


/// "camera" in spherical coordinates
class SphericalCamera
{
  double phi, theta, r; // coordinates
  double dphi, dtheta;  // sphere angle steps
  double rf;            // r increase factor

  // precalculated sines and cosines
  double cos_phi, sin_phi, cos_theta, sin_theta;

public:
  /// initialize step values
  explicit SphericalCamera
    (
      unsigned longitude_steps = 90
    , unsigned latitude_steps  = 50
    , double  distance_factor  = 0.9
    )
      : dphi   (PI_mul_2 / longitude_steps)
      , dtheta (PI / latitude_steps)
      , rf     (distance_factor)
      , phi       (0.0) , theta     (PI_div_2) , r (500.0)
      , cos_phi   (1.0) , sin_phi   (0.0)
      , cos_theta (0.0) , sin_theta (1.0)
  {}

  /// constructor from coordinates
  SphericalCamera
    (
      double phi
    , double theta
    , double r
    , unsigned longitude_steps = 90
    , unsigned latitude_steps  = 50
    , double  distance_factor  = 0.9
    )
      : dphi   (PI_mul_2 / longitude_steps)
      , dtheta (PI / latitude_steps)
      , rf     (distance_factor)
      , phi       (phi)             , theta     (theta)           , r (r)
      , cos_phi   (std::cos(phi))   , sin_phi   (std::sin(phi))
      , cos_theta (std::cos(theta)) , sin_theta (std::sin(theta))
  {}


  /// rotate right
  void right()
  {
    phi += dphi;
    if (PI_mul_2 < phi) phi -= PI_mul_2;

    cos_phi = std::cos(phi);
    sin_phi = std::sin(phi);
  }

  /// rotate left
  void left()
  {
    phi -= dphi;
    if (phi < 0.0) phi += PI_mul_2;

    cos_phi = std::cos(phi);
    sin_phi = std::sin(phi);
  }

  /// rotate up
  void up()
  {
    const double theta1 = theta - dtheta;
    if (0.0 < theta1)
    {
      theta = theta1;
      cos_theta = std::cos(theta);
      sin_theta = std::sin(theta);
    }
  }

  /// rotate down
  void down()
  {
    const double theta1 = theta + dtheta;
    if (theta1 < PI)
    {
      theta = theta1;
      cos_theta = std::cos(theta);
      sin_theta = std::sin(theta);
    }
  }

  /// move forward
  void forward()
  {
    r *= rf;
  }

  /// move backward
  void backward()
  {
    r /= rf;
  }

  /// apply the camera matrix to the projection matrix, which is placed second on the matrix stack
  void apply() const
  {
    glMatrixMode(GL_PROJECTION);

    glPopMatrix();  // duplicate the previous matrix on the stack
    glPushMatrix();
    gluLookAt
      (
        r * sin_theta * cos_phi // camera position x
      , r * sin_theta * sin_phi // camera position y
      , r * cos_theta           // camera position z
      , 0.0, 0.0, 0.0           // camera looks at origin
      , 0.0, 0.0, 1.0           // camera's up vector
      );

    glMatrixMode(GL_MODELVIEW);
  }
};

}
