/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file   Timer.hpp
 * @author Dmitri R. Kuvshinov
 * @brief  timing facility
 *
 * Provides high resolution when possible. 
 * May use Win32 HPET, boost::chrono, C++11 chrono, OpenMP wtime 
 * or std::clock if nothing else is available.
 */
#pragma once


namespace avo
{

/// get current time (starting moment is unspecified and should be considered to be unknown)
double seconds();

/// get timer ticks per second (timer resolution)
double timer_ticks();

/// wait for given period of seconds time (not precise, OS-dependent thread sleep)
void sleep(double);

/// identity functor, doing "nothing" with its arguments
struct Identity
{
  template <class T>       T& operator()(      T &x) const { return x; }
  template <class T> const T& operator()(const T &x) const { return x; }
};


/// timer facility
/**
 * SecondsProxy is a policy for transforming seconds() results if needed
 * RateProxy is a policy for transforming time deltas if needed
 */
template <class RateProxy = Identity, class SecondsProxy = Identity>
class Timer
{
public:
  /// time representation (both period and point)
  typedef double Time;

  /// initializes to the given accumulated time
  explicit Timer
    (
      Time         start_time   = Time(0),
      RateProxy    rateProxy    = RateProxy(),
      SecondsProxy secondsProxy = SecondsProxy()
    )
    : 
      dt(0), 
      st(start_time), 
      rateProxy(rateProxy),
      secondsProxy(secondsProxy)
  {}

  /// reset time to the given accumulated time
  void reset(Time start_time = Time(0))
  {
    dt = Time(0);
    st = start_time;
  }

  /// start measuring time interval
  void start()
  {
    dt = pseconds();
  }

  /// reset and start
  void restart()
  {
    reset();
    start();
  }

  /// stop measuring time interval
  void stop()
  {
    dt = pseconds() - dt;
    st += rateProxy(dt);
  }

  /// accumulate next delta and proceed (eqv. to stop(); start();), returns total
  Time mark()
  {
    const Time t = pseconds();
    st += rateProxy(t - dt);
    dt = t;
    return st;
  }

  /// return delta between last start and this moment, start again
  Time step()
  {
    stop();
    const Time res = dt;
    start();
    return res;
  }

  /// get time elapsed between start() and stop() calls
  Time operator()()
  {
    return dt;
  }

  /// get time elapsed during all intervals between start() stop() calls since reset or creation of the timer
  Time total() const
  {
    return st;
  }

  /// rate proxy accessor
  RateProxy& rate_proxy()
  {
    return rateProxy;
  }

  /// rate proxy const accessor
  const RateProxy& rate_proxy() const
  {
    return rateProxy;
  }

  /// seconds proxy accessor
  SecondsProxy& seconds_proxy()
  {
    return secondsProxy;
  }

  /// seconds proxy const accessor
  const SecondsProxy& seconds_proxy() const
  {
    return secondsProxy;
  }

private:

  Time dt, st; // delta and sum
  RateProxy    rateProxy;
  SecondsProxy secondsProxy;

  Time pseconds()
  {
    return secondsProxy(seconds());
  }
};

}
