/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Vec3.hpp
 * @author Dmitri Kuvshinov
 * @brief 3D vector implementation
 *
 */
#pragma once
#include <cstddef>
#include <cmath>
#include <istream>
#include <ostream>


namespace avo
{

/// 3D vector
template <class Float>
struct Vec3
{
  typedef Float Component;
  typedef Vec3<Float> This;
  Float x, y, z;
  
  explicit Vec3
    (
      Float nx = Float(0.),
      Float ny = Float(0.),
      Float nz = Float(0.)
    )
    : x (nx), y (ny), z (nz) {}

  explicit Vec3(const Float p[3])
    : x (p[0]), y (p[1]), z (p[2]) {}

  /// change sign in-place
  This& chs()
  {
    x = -x;
    y = -y;
    z = -z;
    return *this;
  }

  This operator-() const
  {
    return This(-x, -y, -z);
  }

  This& operator+=(const This &other)
  {
    x += other.x;
    y += other.y;
    z += other.z;
    return *this;
  }

  This& operator-=(const This &other)
  {
    x -= other.x;
    y -= other.y;
    z -= other.z;
    return *this;
  }

  /// by-component multiplication
  This& operator*=(const This &other)
  {
    x *= other.x;
    y *= other.y;
    z *= other.z;
    return *this;
  }

  This& operator*=(const Float coef)
  {
    x *= coef;
    y *= coef;
    z *= coef;
    return *this;
  }

   
  /// by-component division
  This& operator/=(const This &other)
  {
    x /= other.x;
    y /= other.y;
    z /= other.z;
    return *this;
  }

  /// possibly slow, use v *= (1.0 / coef) instead for speed
  This& operator/=(const Float coef)
  {
    x /= coef;
    y /= coef;
    z /= coef;
    return *this;
  }

  /// this dot this, squared length
  Float dotSquare() const
  {
    return x * x + y * y + z * z;
  }

  /// vector length, may overflow for big x, y or z
  Float length() const
  {
    using std::sqrt;
    return sqrt(dotSquare());
  }

  /// normalize this vector in place
  This& normalize()
  {
    const Float len = length();
    if (len != Float(0.))
      *this *= Float(1.) / len;
    return *this;
  }

  /// store per-coordinate minimum
  This& minimize(const This &other)
  {
    if (other.x < x) x = other.x;
    if (other.y < y) y = other.y;
    if (other.z < z) z = other.z;
    return *this;
  }

  /// store per-coordinate maximum
  This& maximize(const This &other)
  {
    if (x < other.x) x = other.x;
    if (y < other.y) y = other.y;
    if (z < other.z) z = other.z;
    return *this;
  }

  /// check if this has greater (or equal) coordinates than the other one
  bool dominates(const This &other) const
  {
    return other.x <= x && other.y <= y && other.z <= z;
  }
};


template <class Float>
inline bool operator==(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return a.x == b.x && a.y == b.y && a.z == b.z;
}

template <class Float>
inline bool operator!=(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return !(a == b);
}


/// lexicographic comparison: x, y, then z
struct LexCompareXYZ
{
  template <class Vec>
  bool operator()(const Vec &a, const Vec &b) const
  {
    return a.x < b.x ||
      (a.x == b.x && (a.y < b.y ||
       (a.y == b.y && a.z < b.z)));
  }
};


/// calculate two vectors sum
template <class Float>
inline Vec3<Float> operator+(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return Vec3<Float>(a.x + b.x, a.y + b.y, a.z + b.z);
}

/// calculate two vectors difference
template <class Float>
inline Vec3<Float> operator-(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return Vec3<Float>(a.x - b.x, a.y - b.y, a.z - b.z);
}

/// by-component vector multiplication
template <class Float>
inline Vec3<Float> operator*(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return Vec3<Float>(a.x * b.x, a.y * b.y, a.z * b.z);
}

/// scale a vector by a coefficient
template <class Float>
inline Vec3<Float> operator*(Float coef, const Vec3<Float> &v)
{
  return Vec3<Float>(coef * v.x, coef * v.y, coef * v.z);
}

/// scale a vector by a coefficient
template <class Float>
inline Vec3<Float> operator*(const Vec3<Float> &v, Float coef)
{
  return coef * v;
}

/// by-component vector division
template <class Float>
inline Vec3<Float> operator/(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return Vec3<Float>(a.x / b.x, a.y / b.y, a.z / b.z);
}

/// possibly slow, use v * (1.0 / coef) instead for speed
template <class Float>
inline Vec3<Float> operator/(const Vec3<Float> &v, Float coef)
{
  return Vec3<Float>(v.x / coef, v.y / coef, v.z / coef);
}


// textual I/O just as number triples

template <class Float>
inline std::ostream& operator<<(std::ostream &os, const Vec3<Float> &v)
{
  return os << v.x << ' ' << v.y << ' ' << v.z;
}

template <class Float>
inline std::istream& operator>>(std::istream &is, Vec3<Float> &v)
{
  return is >> v.x >> v.y >> v.z;
}


/// dot product of two vectors
template <class Float>
inline Float dot(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return a.x * b.x + a.y * b.y + a.z * b.z;
}

/// determinant of [a b ijk] matrix
template <class Float>
inline Vec3<Float> cross(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return Vec3<Float>
    (
      a.y * b.z - a.z * b.y,
      a.z * b.x - a.x * b.z,
      a.x * b.y - a.y * b.x
    );
}

/// compute Euclidian distance
template <class Float>
inline Float distance(const Vec3<Float> &a, const Vec3<Float> &b)
{
  return (a - b).length();
}

/// volume of parallelotop over a, b, c; det[a b c]
template <class Float>
inline double volume(const Vec3<Float> &a, const Vec3<Float> &b, const Vec3<Float> &c)
{
  return dot(cross (a, b), c);
}


/// calculate triangle center (medians intersection point)
template <class Float>
inline Vec3<Float> tricenter(const Vec3<Float> &a, const Vec3<Float> &b, const Vec3<Float> &c)
{
  return Float(1./ 3.) * (a + b + c);
}


/// compute by-coordinate maximima and minima over an array of Vec3
/**
 * n must be > 0
 */
template <class Float>
inline void bbox3(const Vec3<Float> arr[], std::size_t n, Vec3<Float> &pmin, Vec3<Float> &pmax)
{
  if (n == 0)
    return;

  Vec3<Float> cmax = arr[0], cmin = cmax;
  for (std::size_t i = 1; i < n; ++i)
  {
    cmax.maximize (arr[i]);
    cmin.minimize (arr[i]);
  }

  pmin = cmin;
  pmax = cmax;
}


typedef Vec3<float>  Vec3f;
typedef Vec3<double> Vec3d;

}
