﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Pencil
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 *
 * Реализации игры "Жизнь" Конвея в виде построчного сумматора.
 * См. также CellularAutomata.h.
 */
#include <cstdint>
#include <climits>
#include <cstring>
#include <cassert>

#ifdef _MSC_VER
#include <malloc.h>
#include <pmmintrin.h>
#else
#include <x86intrin.h> // (Linux)
#endif

#include <omp.h>

#include "CellularAutomata.hpp"

using namespace std;
using avo::Pixel;

// тип ячейки
typedef uint8_t Cell;
// синоним
typedef IBounded2DCellularAutomaton::Index Index;

// основа-представление состояния игры
struct Bounded2DCellularAutomatonRepresentation
{
  // поле игры "Жизнь" (состояние на момент (old) и счётчики живых соседних клеток (new))
  Cell *cells_old, *cells_new;
  // размеры поля
  Index width, height;
  // ширина линии в памяти
  Index line_stride;
  // отступ от начала выделенного в памяти блока (cells_old, cells_new)
  Index line_offset;

  Bounded2DCellularAutomatonRepresentation()
    : width(0), height(0), line_stride(0), line_offset(0),
      cells_old(nullptr), cells_new(nullptr) {}
};


// базовые реализации стратегий (см. шаблон BasicConwaysLifeLineSummator ниже)

//
// реализация игры "Жизнь" Конвея (Conway's Life)
// см. http://en.wikipedia.org/wiki/Conway%27s_Game_of_Life

// TASK: реализовать какие-нибудь альтернативные правила
// см. http://en.wikipedia.org/wiki/Life-like_cellular_automaton

namespace Basic
{

// простой аллокатор ячеек
struct CellsAllocator
{
  static Cell* alloc_cells(std::size_t amount)
  {
    return new Cell[amount];
  }

  static void dealloc_cells(Cell *cells)
  {
    delete[] cells;
  }
};


struct CellsLineSummator
{
  // затирает строчку нулями
  static void zero_line(Cell *line, const Index stride)
  {
    std::memset(line, 0, stride * sizeof(Cell));
  }

  // Построчные суммы для улучшения локальности доступа
  // строчки [begin, end) из cells_new.
  static void upper_line_sum // записываем сумму верхней строчки
    (const Cell *old_line, Cell *new_line, const Index width)
  {
    // увеличиваем ячейку, если соответствующий пиксель -- не чёрный
    register int left    = old_line[0],
                 middle  = old_line[1];

    // крайняя левая сумма
    new_line[0] = middle + left;

    // "скользящее окно" вдоль горизонтали (аналогично 1D фильтру из Pixels)
    for (Index col = 2; col < width; ++col)
    {
      const register int right = old_line[col];
      new_line[col - 1] = left + middle + right;
      left = middle;
      middle = right;
    }

    // крайняя правая сумма
    new_line[width - 1] = left + middle;
  }

  static void middle_line_sum // добавляем суммы средней строчки (слева + справа)
    (const Cell *old_line, Cell *new_line, const Index width)
  {
    register int left    = old_line[0],
                 middle  = old_line[1];

    new_line[0] += middle;
    for (Index col = 2; col < width; ++col)
    {
      const register int right = old_line[col];
      new_line[col - 1] += right + left;
      left = middle;
      middle = right;
    }
    new_line[width - 1] += left;
  }

  static void lower_line_sum // добавляем суммы нижней строчки
    (const Cell *old_line, Cell *new_line, const Index width)
  {
    register int left    = old_line[0],
                 middle  = old_line[1];

    new_line[0] += middle + left;
    for (Index col = 2; col < width; ++col)
    {
      const register int right = old_line[col];
      new_line[col - 1] += left + middle + right;
      left = middle;
      middle = right;
    }
    new_line[width - 1] += left + middle;
  }
};


struct CellsFromPicture
{
  // построить исходное поле на основе 32-битной картинки (чёрный => 0, нечёрный => 1)
  static void from_pixels_line(const Pixel *src_line, Cell *dst_line, const Index width)
  {
    for (Index i = 0; i < width; ++i)
      dst_line[i] = src_line[i].binary != 0; // 0 or 1
  }
};


struct CellsFinalizer
{
  // привести cells_new к значениям 0, 1 в соответствии с правилами, записать картинку в dst
  static void cells_new_finalize
    (Bounded2DCellularAutomatonRepresentation &repr, Index begin, Index end, Pixel *dst)
  {
    const Index width = repr.width,
                line_stride = repr.line_stride,
                offset = begin * line_stride;

    const Cell  *old_line    = repr.cells_old + offset;
          Cell  *new_line    = repr.cells_new + offset;
          Pixel *pixels_line = dst            + offset;

    for (Index line = begin; line < end; ++line)
    {
      for (Index i = 0; i < width; ++i)
      {
        register int new_value = old_line[i] != 0;
        if (new_value) // клетка погибает, если имеет < 2 или > 3 соседей
        {
          if (new_line[i] < 2 || new_line[i] > 3)
            new_value = 0;
        }
        else if (new_line[i] == 3)  // место с 3 соседними клетками занимается новой клеткой
        {
          new_value = 1;
        }
  
        new_line[i] = new_value;
        pixels_line[i].binary = new_value? ~avo::PixelInt(0): avo::PixelInt(0);
      }

      old_line += line_stride;
      new_line += line_stride;
      pixels_line += width;
    }
  }
};

}



// Обобщённая реализация.
// Определения по умолчанию задают "обычную" реализацию.
// Детали реализации передаются в виде стратегий (термин А.Александреску):
// CellsAllocator   -- создание и удаление массивов ячеек;
// CellsLineSummator  -- суммирование линий;
// CellsFromPicture -- принятие картинки;
// CellsFinalizer   -- применение правил и формирование картинки.
template
<
  class CellsAllocator    = Basic::CellsAllocator,
  class CellsLineSummator = Basic::CellsLineSummator,
  class CellsFromPicture  = Basic::CellsFromPicture,
  class CellsFinalizer    = Basic::CellsFinalizer
>
class BasicConwaysLifeLineSummator
  : public IBounded2DCellularAutomaton,
    protected Bounded2DCellularAutomatonRepresentation,
    private CellsAllocator,
    private CellsLineSummator,
    private CellsFromPicture,
    private CellsFinalizer
{
public:
  // конструктор по умолчанию пустой (работают конструкторы базовых классов)
  BasicConwaysLifeLineSummator() {}
  // деструктор
  ~BasicConwaysLifeLineSummator() override
  {
    cleanup();  
  }

  // определение функции интерфейса, здесь stride == w
  void reshape(Index w, Index h) override
  {
    reshape(w, h, w);
  }

  // название автомата
  const char* name() override
  {
    return "Conway's Life: line summator";
  }

  // получить текущие размеры поля
  Extents extents() const override
  {
    return Extents(width, height);
  }

  // построить исходное поле на основе 32-битной картинки (чёрный => 0, нечёрный => 1)
  void fromPixels(const Pixel *src) override
  {
    #pragma omp parallel for
    for (Index line = 0; line < height; ++line)
      from_pixels_line(src + line * width, cells_old + line_offset + line * line_stride, width);
  }

  // следующий шаг по времени, сформировать 32-битную картинку в dst
  void next(Pixel *dst) override
  {
    // параллельно по строчкам (создаёт группу потоков)
    #pragma omp parallel
    {
      const int tid = omp_get_thread_num(),  // id текущего потока из [0, threads)
            threads = omp_get_num_threads(); // количество потоков в группе

      // строчки в cells, принадлежащие текущему потоку
      const Index begin = (tid * height) / threads,
                  end  = ((tid + 1) * height) / threads;

      const Index width  = this->width,
                  stride = this->line_stride,
                  offset = begin * stride + line_offset;
      
      // текущие строки
      const Cell *cur_line  = cells_old + offset,
                 *prev_line = cur_line - stride,
                 *next_line = cur_line + stride;

      Cell *new_line = cells_new + offset;

      // первая строчка результата
      if (tid == 0)
        zero_line (new_line, stride);
      else
        upper_line_sum (prev_line, new_line, width);

      middle_line_sum (cur_line, new_line, width);
      lower_line_sum (next_line, new_line, width);

      for (Index lines = begin + 2; lines < end; ++lines)
      {
        prev_line =  cur_line;
        cur_line  =  next_line;
        next_line += stride;
        new_line  += stride;

        // средние строчки: 3 смещения по строчкам
        upper_line_sum (prev_line, new_line, width);
        middle_line_sum (cur_line, new_line, width);
        lower_line_sum (next_line, new_line, width);
        // отличие upper_line_sum от lower_line_sum в том,
        // что она затирает старые значения в new_line, а не добавляет к ним свои суммы
      }

      new_line += stride;
      // последняя строчка результата
      upper_line_sum  (cur_line, new_line, width);
      middle_line_sum (next_line, new_line, width);
      if (tid != threads - 1)
        lower_line_sum (next_line + stride, new_line, width);

      // получить новые значения pixels, в соответствии с правилом "Жизнь"
      cells_new_finalize(*this, begin, end, dst);
    }

    // обменять cells_new и cells_old
    swap(cells_new, cells_old);
  }

protected:

  // обновить размеры поля
  void reshape(Index w, Index h, Index stride)
  {
    if (w < 32) w = 32;
    if (h < 4)  h = 4;
    if (stride < w) stride = w;
    
    if (w != width || h != height || stride != line_stride)
    {
      cleanup();
      width = w;
      height = h;
      line_stride = stride;
      const size_t sz = static_cast<size_t>(stride) * h;

      cells_old = alloc_cells(sz);
      std::memset(cells_old, 0, sz * sizeof(Cell));

      cells_new = alloc_cells(sz);
      std::memset(cells_new, 0, sz * sizeof(Cell)); // возможно, избыточно
    }
  }

private:

  // уничтожает хранимые ресурсы
  void cleanup()
  {
    dealloc_cells(cells_old);
    dealloc_cells(cells_new);
  }

  // копирование запрещено
  typedef BasicConwaysLifeLineSummator
    <CellsAllocator, CellsLineSummator, CellsFromPicture, CellsFinalizer> This;

  BasicConwaysLifeLineSummator(const This&);
  This& operator=(const This&);
};



///////////////////////////////////////////////////////////////////////////////////////////////////
// SSE3-вариант

namespace SSE3
{

#ifndef _MSC_VER
#error The following code uses _aligned_malloc and _aligned_free provided by MSVC, substitute variants for your compiler and disable this error
#endif

// выровненный аллокатор ячеек, MSVC
struct AlignedCellsAllocator
{
  static Cell* alloc_cells(std::size_t amount)
  {
    return static_cast<Cell*>(_aligned_malloc(amount * sizeof(Cell), 16));
  }

  static void dealloc_cells(Cell *cells)
  {
    _aligned_free(cells);
  }
};


struct CellsLineSummator
{
  static_assert (sizeof(Cell) * CHAR_BIT == 8, "CellsLineSummatorSSE3: cells are supposed to be 8-bit bytes");
   
  // затирает строчку нулями
  static void zero_line(Cell *line, const Index stride)
  {
    // строчка выровнена по 16
    const Index xmm_stride = stride >> 4;
    __m128i *xmm_line = (__m128i*)line,
            *xmm_line_end = xmm_line + (xmm_stride - 2);

    const __m128i zero = _mm_setzero_si128();
    // Чтобы выполнять в цикле 4 записи (64 байта == строка кэша) за итерацию,
    // следует сначала выровнять по 64 (при этом помним, что линия уже выровнена по 16).
    switch (xmm_stride & 3)
    {
    case 3: _mm_store_si128(xmm_line++, zero);
    case 2: _mm_store_si128(xmm_line++, zero);
    case 1: _mm_store_si128(xmm_line++, zero);
    }
    // сам цикл
    // (на Intel Sandy/Ivy Bridge простой цикл с одной записью за итерацию мб эффективнее)
    for (xmm_line += 2; xmm_line <= xmm_line_end; xmm_line += 4)
    {
      _mm_store_si128(xmm_line - 2, zero);
      _mm_store_si128(xmm_line - 1, zero);
      _mm_store_si128(xmm_line + 0, zero);
      _mm_store_si128(xmm_line + 1, zero);
    }
  }


  // Построчные суммы для улучшения локальности доступа
  // строчки [begin, end) из cells_new.
  static void upper_line_sum // записываем сумму верхней строчки
    (const Cell *old_line, Cell *new_line, const Index width)
  {
    for (Index col = 0; col < width; col += 16)
    {
      const __m128i oldv1 = _mm_load_si128( (const __m128i*)old_line ),
                    oldv0 = _mm_lddqu_si128( (const __m128i*)(old_line - 1) ),
                    sum0  = _mm_add_epi8( oldv0, oldv1 ),
                    oldv2 = _mm_lddqu_si128( (const __m128i*)(old_line + 1) );

      _mm_store_si128( (__m128i*)new_line, _mm_add_epi8( oldv2, sum0 ) );
      old_line += 16;
      new_line += 16;
    }
  }

  static void middle_line_sum // добавляем суммы средней строчки (слева + справа)
    (const Cell *old_line, Cell *new_line, const Index width)
  {
    for (Index col = 0; col < width; col += 16)
    {
      const __m128i oldv0 = _mm_lddqu_si128( (const __m128i*)(old_line - 1) ),
                    oldv2 = _mm_lddqu_si128( (const __m128i*)(old_line + 1) ),
                    newv  = _mm_load_si128( (const __m128i*)new_line ),
                    sum1  = _mm_add_epi8( newv, oldv2 );

      _mm_store_si128( (__m128i*)new_line, _mm_add_epi8( sum1, oldv0 ) );
      old_line += 16;
      new_line += 16;
    }
  }

  static void lower_line_sum // добавляем суммы нижней строчки
    (const Cell *old_line, Cell *new_line, const Index width)
  {
    for (Index col = 0; col < width; col += 16)
    {
      const __m128i oldv1 = _mm_load_si128( (const __m128i*)old_line ),
                    oldv0 = _mm_lddqu_si128( (const __m128i*)(old_line - 1) ),
                    sum0  = _mm_add_epi8( oldv0, oldv1 ),
                    oldv2 = _mm_lddqu_si128( (const __m128i*)(old_line + 1) ),
                    newv  = _mm_load_si128( (const __m128i*)new_line ),
                    sum1  = _mm_add_epi8( newv, oldv2 );

      _mm_store_si128( (__m128i*)new_line, _mm_add_epi8( sum1, sum0 ) );
      old_line += 16;
      new_line += 16;
    }
  }
};


// В общем-то просто упражнение в использовании SIMD, поскольку это редко используемая операция
// (часто вызывается разве что при рисовании мышью одновременно с работой клеточного автомата).
struct CellsFromPicture
{
  static_assert (sizeof(Pixel) * CHAR_BIT == 32, "CellsLineSummatorSSE3: pixels are supposed to be 32-bit chunks");
  
  // Считать одну строчку пикселей и заполнить одну строчку клеток,
  // по клеткам имеем выравнивание с отступами с обоих концов, по пикселям такого нет.
  static void from_pixels_line(const Pixel *src_line, Cell *dst_line, const Index width)
  {
    // 4 пикселя = 16 байт = регистр SSE -> 4 ячейки
    // 16 байт = 16 ячеек  = регистр SSE <- 16 пикселей = 64 байта
    // цикл обрабатывает по 16 пикселей
    const __m128i *src16 = (const __m128i*)src_line + 2;
          __m128i *dst16 =       (__m128i*)dst_line;

    const __m128i zero = _mm_setzero_si128(),
              all_ones = _mm_set_epi32(0x01010101, 0x01010101, 0x01010101, 0x01010101);

    // перестраховка: затереть ведущий отступ
    _mm_store_si128(dst16 - 1, zero);

    for (Index w16 = width >> 4; w16 != 0; --w16)
    {
      const __m128i p0 = _mm_lddqu_si128(src16 - 2),
                    p4 = _mm_lddqu_si128(src16 - 1),
                    p8 = _mm_lddqu_si128(src16 + 0),
                    pC = _mm_lddqu_si128(src16 + 1);
      src16 += 4;
      const __m128i p0_7 = _mm_packs_epi32(p0, p4),
                    p8_F = _mm_packs_epi32(p8, pC);

      // предполагает, что пиксели были либо чёрные с нулевым альфа, либо белые с альфа==255!
      const __m128i cells = _mm_and_si128( all_ones, _mm_packs_epi16(p0_7, p8_F) );
      _mm_store_si128(dst16++, cells);
    }

    // затереть остаток
    _mm_store_si128(dst16, zero);

    // добить остаток (до 15 пикселей)
    for (Index i = width & -16; i < width; ++i)
      dst_line[i] = src_line[i].binary != 0;
  }
};


struct CellsFinalizer
{
  // привести cells_new к значениям 0, 1 в соответствии с правилами, записать картинку в dst
  static void cells_new_finalize
    (Bounded2DCellularAutomatonRepresentation &repr, Index begin, Index end, Pixel *dst)
  {
    const __m128i all_ones   = _mm_set_epi32(0x01010101, 0x01010101, 0x01010101, 0x01010101),
                  shift_mask = _mm_set_epi32(0x7F7F7F7F, 0x7F7F7F7F, 0x7F7F7F7F, 0x7F7F7F7F),
                  pixel_mask = _mm_set_epi32(0xFF, 0xFF, 0xFF, 0xFF);

    const Index width  = repr.width,
                stride = repr.line_stride,
                offset = stride * begin + repr.line_offset;

    const Cell *old_line = repr.cells_old + offset;
          Cell *new_line = repr.cells_new + offset;
         Pixel *pixels_line = dst + width * begin;

    while (begin++ != end)
    {
      for (Index col = 0, max_col = width & -16; col < max_col; col += 16)
      {
                __m128i *new_p =       (__m128i*)(new_line + col);
          const __m128i *old_p = (const __m128i*)(old_line + col);

          // переделанное правило игры Жизнь для удобства SIMD-записи,
          // более удобочитаемая версия ниже (обработка остатка строки)
          const __m128i new_vec            = _mm_load_si128( new_p ),
                        old_vec            = _mm_load_si128( old_p ),
                        shifted_new_vec    = _mm_and_si128( _mm_srli_epi16( new_vec, 1 ), shift_mask ),
                        masked_old_new_vec = _mm_or_si128( old_vec, _mm_and_si128( new_vec, all_ones ) ),
                        if_ones_1          = _mm_cmpeq_epi8( shifted_new_vec, all_ones ),
                        if_ones_2          = _mm_cmpeq_epi8( masked_old_new_vec, all_ones ),
                        test               = _mm_and_si128( if_ones_1, if_ones_2 );

          // запись cells_new
          _mm_store_si128(new_p, _mm_and_si128(test, all_ones));

          // запись картинки
                __m128i *pixels_p     = (__m128i*)(pixels_line + col);
          const __m128i zero          = _mm_setzero_si128(),
                        test01234567  = _mm_unpacklo_epi8( test, zero ), // "удвоить битность", перемежая нулями (8 -> 16)
                        test0123      = _mm_unpacklo_epi16( test01234567, zero ), // ещё раз (16 -> 32)
                        test4567      = _mm_unpackhi_epi16( test01234567, zero );
          
          // строчки пикселей (куда пишем) невыровнены
          _mm_storeu_si128( pixels_p + 0, _mm_cmpeq_epi32( test0123, pixel_mask ) );
          _mm_storeu_si128( pixels_p + 1, _mm_cmpeq_epi32( test4567, pixel_mask ) );

          const __m128i test89ABCDEF  = _mm_unpackhi_epi8( test, zero ), // другая половина
                        test89AB      = _mm_unpacklo_epi16( test89ABCDEF, zero ),
                        testCDEF      = _mm_unpackhi_epi16( test89ABCDEF, zero );
      
          _mm_storeu_si128( pixels_p + 2, _mm_cmpeq_epi32( test89AB, pixel_mask ) );
          _mm_storeu_si128( pixels_p + 3, _mm_cmpeq_epi32( testCDEF, pixel_mask ) );
      }

      // остатотк строки
      for (Index col = width & -16; col < width; ++col)
      {
        // переделанное правило игры Жизнь
        if ((old_line[col] | (new_line[col] & 1)) == 1 && (new_line[col] >> 1) == 1)
        {
          pixels_line[col].binary = ~avo::PixelInt(0);
          new_line[col] = 1;
        }
        else
        {
          pixels_line[col].binary = 0;
          new_line[col] = 0;
        }
      }

      old_line += stride;
      new_line += stride;
      pixels_line += width;
    }
  }
};

}


typedef
  BasicConwaysLifeLineSummator
  <
    SSE3::AlignedCellsAllocator,
    SSE3::CellsLineSummator,
    SSE3::CellsFromPicture,
    SSE3::CellsFinalizer
  >
    ConwaysLifeLineSummatorSSE3Base;

class ConwaysLifeLineSummatorSSE3 : public ConwaysLifeLineSummatorSSE3Base
{
public:
  ConwaysLifeLineSummatorSSE3()
  {
    line_offset = 16;
  }

  // название автомата
  const char* name() override
  {
    return "Conway's Life: line summator/SSE3";
  }

  void reshape(Index w, Index h) override
  {
    // требуется, чтобы все строчки были выровнены по 16,
    // кроме того, чтобы не добавлять в операцию суммирования обработку специальных случаев
    // на границе строк, добавляются отступы слева и справа от строки (до 32 байт в сумме)
    ConwaysLifeLineSummatorSSE3Base::reshape(w, h, (w + 32) & -16);
  }
};


///////////////////////////////////////////////////////////////////////////////////////////////////

// вычисляет построчные суммы, использует OpenMP для распараллеливания
IBounded2DCellularAutomaton* newConwaysLifeLineSummator()
{
  return new BasicConwaysLifeLineSummator<>;
}

// работает аналогично , но использует SSE3
IBounded2DCellularAutomaton* newConwaysLifeLineSummatorSSE3()
{
  return new ConwaysLifeLineSummatorSSE3;
}
