﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Pencil
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * В примере пользователь может рисовать белым "карандашом" на чёрном "холсте"
 * с помощью мыши (зажимая левую кнопку).
 * Кроме того, есть дополнительные "опции" ;)
 *
 * Нажатие
 *  ENTER/SPACE -- очистка экрана;
 *  ESC/Q -- выход из программы;
 *  стрелки -- прокручивание "холста";
 *  R -- набросать псевдослучайных точек;
 *  M -- изобразить псевдослучайный "лабиринт";
 *  L -- старт/стоп игру "Жизнь" Конвея;
 *  ; -- выбрать предыдущий автомат из списка;
 *  ' -- выбрать следующий автомат из списка;
 *  [ -- уменьшить число потоков OpenMP;
 *  ] -- увеличить число потоков OpenMP.
 */
#include <GL/freeglut.h>
#include <cstdlib>  // rand/srand
#include <cstring>  // memset, memcpy, memmove
#include <cmath>    // abs
#include <iostream>
#include <algorithm> // max, min
#include <omp.h> // OpenMP

#include <avo/Pixel.hpp> // RGBA-цвет в 32-битном формате (8-битные компоненты типа uint8_t)
#include <avo/Timer.hpp>
#include "CellularAutomata.hpp"

using avo::Pixel;
using avo::Timer;

//
// константы и глобальные переменные
//

/// счётчик кадров (сбрасывается каждые 10с)
int frames;
/// момент времени, с которого происходит накопление счётчика кадров
double frames_reported_last;

/// текущий размер (клиентской области) окна в пикселях
int width = 800, height = 600;
/// поправка координат мыши (координаты левого верхнего угла клиентской области окна в окне)
int mouse_origin_x, mouse_origin_y;

/// массив пикселей размера width * height
Pixel *pixels;

/// цвет карандаша (белый по умолчанию)
Pixel pencil = Pixel::grey(255, 255);

/// последние полученные координаты указателя мыши
int mouse_x, mouse_y;

/// признак "рисование" (зажата левая кнопка мыши)
bool draw;

/// размер ячейки псевдослучайного лабиринта в пикселях (см. simple_maze())
const int maze_width = 20, maze_height = 20;

// состояние игры "Жизнь"
// признак "игра Жизнь запущена"
bool life_game_running;
const int automata_size = 3;
int automaton;
IBounded2DCellularAutomaton *automata[automata_size];


//
// вспомогательные функции
//

/// создать новый чёрный холст размера w * h пикселей
Pixel* black_canvas(int w, int h)
{
  Pixel *canvas = new Pixel[w * (h + 1)]; // дополнительная строчка используется при прокручивании вверх/вниз
  // 0 даёт чёрный цвет, можно получить оттенок серого, выбрав значение из диапазона [0, 255]
  std::memset(canvas, 0, w * h * sizeof(Pixel));
  return canvas;
}


/// заменить текущую картинку чёрным холстом (сбросить)
void black_canvas()
{
  delete[] pixels;
  pixels = black_canvas(width, height);
}

/// нарисовать отрезок прямой, соединяющий пиксели (x0, y0) -- (x1, y1) (алгоритм Брезенхема)
void line(int x0, int y0, int x1, int y1) // см. http://en.wikipedia.org/wiki/Bresenham%27s_algorithm
{
  const int dx = std::abs(x1 - x0),
            dy = std::abs(y1 - y0),
            sx = x0 < x1? 1: -1,
            sy = y0 < y1? 1: -1;

  int error = dx - dy;
  while (true)
  {
    pixels[x0 + y0 * width] = pencil;
    if (x0 == x1 && y0 == y1)
      break;

    const int error_twice = 2 * error;
    if (error_twice > -dy)
      error -= dy, x0 += sx;
    if (error_twice < +dx)
      error += dx, y0 += sy;
  }
}

// Реализация line выше не выполняет проверку выхода за пределы pixels,
// поэтому возможен крах программы при рисовании линий по координатам, выходящим за пределы картинки.
// Можно дописать защищённый вариант line, выполняющий отсечение линии по границам картинки.


//
// операции
//

/// набросать случайных точек
void random_dots()
{
  const int size = width * height;
  for (int i = 0; i < size; ++i)
    if (256 * std::rand() < RAND_MAX)
      pixels[i] = pencil;
}


/// нарисовать псевдослучайный лабиринт (алгоритм "10 PRINT", см. http://10print.org/ )
void simple_maze()
{
  for (int x = maze_width; x < width; x += maze_width)
    for (int y = maze_height; y < height; y += maze_height)
      std::rand() <= RAND_MAX / 2 ?  // "стиль" специально слегка "подпорчен" :)
        line(x - maze_width, y - maze_height, x, y):
        line(x, y - maze_height, x - maze_width, y);
}

// TASK: реализовать отрисовку какого-нибудь сравнительно сложного узора,
// например, "круглого" псевдослучайного лабиринта в центре.


//
// игра "Жизнь" меняет привязку idle-функции GLUT (вызывается из главного цикла при отсутствии сообщений)

/// сформировать cells_old на основе pixels
void conways_life_from_pixels()
{
  if (life_game_running)
  {
    automata[automaton]->reshape(width, height);
    automata[automaton]->fromPixels(pixels);
  }
}

void conways_life_step();
/// инициализация и привязка
void conways_life_start()
{
  conways_life_from_pixels();
  glutIdleFunc(conways_life_step);
}

/// останов
void conways_life_stop()
{
  glutIdleFunc(nullptr);
}

/// алгоритм выполняет суммирование "живых" соседей ячеек в cells и заполняет pixels соответственно правилам "Жизни"
void conways_life_step()
{
  // проверим, не изменились ли размеры окна (и, соответственно, пОля "Жизни")
  const auto extents = automata[automaton]->extents();
  if (width != extents.first || height != extents.second)
  {
    // выполним перезапуск
    conways_life_stop();
    conways_life_start();
  }

  // собственно вычисление следующего состояния
  automata[automaton]->next(pixels);

  // показать
  glutPostRedisplay();
}


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  if (!pixels) return;
  glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
  glFlush();
  
  // счётчик кадров в секунду, выводит среднее за каждые 10с
  ++frames;
  const double dt = avo::seconds() - frames_reported_last;
  if (dt >= 10.0)
  {
    std::cout << frames / dt << '\n';
    frames = 0;
    frames_reported_last += dt;
  }
  
  glFinish();
}


/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  // получить положение клиентской области окна
  mouse_origin_x = std::max(0, glutGet(GLUT_WINDOW_X));
  mouse_origin_y = std::max(0, glutGet(GLUT_WINDOW_Y));

  // размеры изменились и корректны? (вторая проверка избыточна на практике)
  if ((w != width || h != height) && (w > 0 && h > 0))
  {
    Pixel *new_pixels = black_canvas(w, h);
    if (pixels) // есть, что копировать
    {
      // копируем прямоугольник, вписывающийся в новые координаты (с левого нижнего угла)
      const int min_w = std::min(w, width),
                min_h = std::min(h, height),
                min_line_size = min_w * sizeof(Pixel); // размер копируемой строки в байтах

      for (int line = 0; line < min_h; ++line)
      {
        std::memcpy(new_pixels + line * w,     // место назначения
                    pixels     + line * width, // копируемые данные
                    min_line_size);            // размер блока в байтах
      }
    }

    // заменим старое содержимое pixels на новое
    delete[] pixels;
    pixels = new_pixels;

    // установим новые размеры картинки
    width = w;
    height = h;
    if (mouse_x >= width)
      mouse_x = width - 1;
    if (mouse_y >= height)
      mouse_y = height - 1;
    
    // установить новые координаты окна просмотра (вся клиентская область)
    glViewport(0, 0, w, h);
  } 
}


/// вывести в консоль название текущего автомата
void report_current_automaton()
{
  std::cout << "\nAutomaton: " << automata[automaton]->name() << std::endl;
}

/// вывести в консоль количество потоков в рабочей группе OpenMP
void report_current_threads()
{
  std::cout << "\nThreads: " << omp_get_max_threads() << std::endl;
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case ' ': 
  case '\r': case '\n':  black_canvas(); conways_life_from_pixels(); break;
  case 'r':  case 'R':   random_dots();  conways_life_from_pixels(); break;
  case 'm':  case 'M':   simple_maze();  conways_life_from_pixels(); break;
  case 'l':  case 'L':
    // стартовать или остановить игру "Жизнь"
    if (life_game_running = !life_game_running) // изменяет значение признака на противоположное
      conways_life_start();
    else
      conways_life_stop();
    break;

  case ';': // предыдущая реализация в automata
    if (automaton == 0) automaton = automata_size;
    --automaton;
    report_current_automaton();
    conways_life_from_pixels();
    break;

  case '\'': // следующая реализация в automata
    if (++automaton == automata_size) automaton = 0;
    report_current_automaton();
    conways_life_from_pixels();
    break;

  case '[': // убрать поток из рабочей группы OpenMP
    {
      const int threads = omp_get_max_threads();
      if (threads > 1)
        omp_set_num_threads(threads - 1);
      report_current_threads();
    }
    break;

  case ']': // добавить поток в рабочую группу OpenMP
    {
      const int threads = omp_get_max_threads();
      if (threads < omp_get_num_procs())
        omp_set_num_threads(threads + 1);
      report_current_threads();
    }
    break;

  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    return;

  default:
    return; // другие клавиши игнорируются
  }
  glutPostRedisplay();
}


/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_special_key(int key, int x, int y)
{
  // размер строчки пикселей в байтах (используется при прокручивании картинки)
  const int line_size = width * sizeof(Pixel);
  // нажатие на стрелки прокручивает картинку
  switch (key)
  {
  case GLUT_KEY_LEFT:
    for (int line = 0; line < height; ++line)
    {
      Pixel* base = pixels + line * width;
      Pixel stored = base[0];
      std::memmove(base, base + 1, line_size - sizeof(Pixel));
      base[width - 1] = stored;
    }
    conways_life_from_pixels();
    break;

  case GLUT_KEY_UP:
    // используем дополнительную строчку в конце массива pixels
    {
      Pixel* extra_line = pixels + width * height;
      std::memcpy(extra_line, extra_line - width, line_size);
      std::memmove(pixels + width, pixels, line_size * (height - 1));
      std::memcpy(pixels, extra_line, line_size);
    }
    conways_life_from_pixels();
    break;

  case GLUT_KEY_RIGHT:
    for (int line = 0; line < height; ++line)
    {
      Pixel* base = pixels + line * width;
      Pixel stored = base[width - 1];
      std::memmove(base + 1, base, line_size - sizeof(Pixel));
      base[0] = stored;
    }
    conways_life_from_pixels();
    break;

  case GLUT_KEY_DOWN:
    // используем дополнительную строчку в конце массива pixels
    {
      Pixel* extra_line = pixels + width * height;
      std::memcpy(extra_line, pixels, line_size);
      std::memmove(pixels, pixels + width, line_size * (height - 1));
      std::memcpy(extra_line - width, extra_line, line_size);
    }
    conways_life_from_pixels();
    break;

  default:
    return; // другие клавиши игнорируются
  }
  glutPostRedisplay();
}



// проверка качества координат мыши
inline bool good_coords(int x, int y)
{
  return x >= 0 && y >= 0 && x < width && y < height;
}


/// клик мыши, рисуем точку по левому клику
void mouse_click(int button, int state, int x, int y)
{
  y = height - y; // координаты мыши отсчитываются с верхнего левого угла
  if (button != GLUT_LEFT_BUTTON || !good_coords(x, y))
    return;

  if (draw = state == GLUT_DOWN) // нажали?
  {
    // сохранить текущие координаты (для отрисовки линии)
    mouse_x = x;
    mouse_y = y;
    // изобразить точку
    pixels[x + y * width] = pencil;
    conways_life_from_pixels();
    glutPostRedisplay();
  }
}


/// отработать движение мыши (рисует отрезок (mouse_x, mouse_y) -- (x, y), если зажата левая кнопка)
void mouse_move(int x, int y)
{
  y = height - y; // координаты мыши отсчитываются с верхнего левого угла
  if (draw && good_coords(x, y))
  {
    line(mouse_x, mouse_y, x, y);
    mouse_x = x;
    mouse_y = y;
    conways_life_from_pixels();
    glutPostRedisplay();
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);
  
  // инициализировать pixels на случай невызова reshape на старте (так происходит в сборке MSVC2010)
  black_canvas();
  automata[0] = newConwaysLifeCellwiseSummator();
  automata[1] = newConwaysLifeLineSummator();
  automata[2] = newConwaysLifeLineSummatorSSE3();
  
  report_current_threads();
  report_current_automaton();
  
  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_SINGLE); // режим: 32-битный цвет RGBA, одинарный буфер кадра SINGLE
  glutInitWindowPosition(0, 0);       // координаты верхнего левого угла окна на экране
  glutInitWindowSize(width, height);  // полные ширина и высота окна
  glutCreateWindow("Pencil");         // создать окно, в функцию передаётся заголовок создаваемого окна

  // привязка функций-обработчиков событий
  glutDisplayFunc(repaint);             // перерисовка
  glutReshapeFunc(reshape);             // изменение размеров окна
  glutKeyboardFunc(handle_key);         // обработка нажатий "обычных" клавиш
  glutSpecialFunc(handle_special_key);  // обработка нажатий "специальных" клавиш
  glutMouseFunc(mouse_click);           // обработка нажатий кнопок мыши
  glutMotionFunc(mouse_move);           // обработка перемещения мыши
  // сменить указатель мыши на крест
  glutSetCursor(GLUT_CURSOR_CROSSHAIR);

  // старт
  frames_reported_last = avo::seconds();
  std::cout << "Starting Pencil. FPS:\n";
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
