﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Pencil
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * В примере пользователь может рисовать белым "карандашом" на чёрном "холсте"
 * с помощью мыши (зажимая левую кнопку).
 * Кроме того, есть дополнительные "опции" ;)
 *
 * Нажатие
 *  ENTER/SPACE -- очистка экрана,
 *  ESC/Q -- выход из программы,
 *  стрелки -- прокручивание "холста", 
 *  R -- набросать псевдослучайных точек,
 *  M -- изобразить псевдослучайный "лабиринт",
 *  L -- старт/стоп игру "Жизнь" Конвея.
 */
#include <GL/freeglut.h>
#include <cstdlib>  // rand/srand
#include <cstring>  // memset, memcpy, memmove
#include <cmath>    // abs
#include <iostream>
#include <algorithm> // только ради std::min, std::max

#include <avo/Pixel.hpp> // RGBA-цвет в 32-битном формате (8-битные компоненты типа uint8_t)
#include <avo/Timer.hpp>

using avo::Pixel;
using avo::Timer;

//
// константы и глобальные переменные
//

/// счётчик кадров (сбрасывается каждые 10с)
int frames;
/// момент времени, с которого происходит накопление счётчика кадров
double frames_reported_last;

/// текущий размер (клиентской области) окна в пикселях
int width = 800, height = 600;
/// поправка координат мыши (координаты левого верхнего угла клиентской области окна в окне)
int mouse_origin_x, mouse_origin_y;

/// массив пикселей размера width * height
Pixel *pixels;

/// цвет карандаша (белый по умолчанию)
Pixel pencil = Pixel::grey(255);

/// последние полученные координаты указателя мыши
int mouse_x, mouse_y;

/// признак "рисование" (зажата левая кнопка мыши)
bool draw;

/// размер ячейки псевдослучайного лабиринта в пикселях (см. simple_maze())
const int maze_width = 20, maze_height = 20;

// состояние игры "Жизнь"
// признак "игра Жизнь запущена"
bool life_game_running;
/// размер поля игры "Жизнь" (в ячейках, каждая ячейка отвечает одному пикселю)
int life_game_width, life_game_height;

/// тип ячейки
typedef int Cell;
/// поле игры "Жизнь"
Cell *cells;


//
// вспомогательные функции
//

/// создать новый чёрный холст размера w * h пикселей
Pixel* black_canvas(int w, int h)
{
  Pixel *canvas = new Pixel[w * (h + 1)]; // дополнительная строчка используется при прокручивании вверх/вниз
  // 0 даёт чёрный цвет, можно получить оттенок серого, выбрав значение из диапазона [0, 255]
  std::memset(canvas, 0, w * h * sizeof(Pixel));
  return canvas;
}


/// заменить текущую картинку чёрным холстом (сбросить)
void black_canvas()
{
  delete[] pixels;
  pixels = black_canvas(width, height);
}

/// нарисовать отрезок прямой, соединяющий пиксели (x0, y0) -- (x1, y1) (алгоритм Брезенхема)
void line(int x0, int y0, int x1, int y1) // см. http://en.wikipedia.org/wiki/Bresenham%27s_algorithm
{
  const int dx = std::abs(x1 - x0),
            dy = std::abs(y1 - y0),
            sx = x0 < x1? 1: -1,
            sy = y0 < y1? 1: -1;

  int error = dx - dy;
  while (true)
  {
    pixels[x0 + y0 * width] = pencil;
    if (x0 == x1 && y0 == y1)
      break;

    const int error_twice = 2 * error;
    if (error_twice > -dy)
      error -= dy, x0 += sx;
    if (error_twice < +dx)
      error += dx, y0 += sy;
  }
}


//
// операции
//

/// набросать случайных точек
void random_dots()
{
  const int size = width * height;
  for (int i = 0; i < size; ++i)
    if (256 * std::rand() < RAND_MAX)
      pixels[i] = pencil;
}


/// нарисовать псевдослучайный лабиринт (алгоритм "10 PRINT", см. http://10print.org/ )
void simple_maze()
{
  for (int x = maze_width; x < width; x += maze_width)
    for (int y = maze_height; y < height; y += maze_height)
      std::rand() <= RAND_MAX / 2 ?  // "стиль" специально слегка "подпорчен" :)
        line(x - maze_width, y - maze_height, x, y):
        line(x, y - maze_height, x - maze_width, y);
}

// TASK: реализовать отрисовку какого-нибудь сравнительно сложного узора,
// например, "круглого" псевдослучайного лабиринта в центре.


//
// реализация игры "Жизнь" Конвея (Conway's Life)
// см. http://en.wikipedia.org/wiki/Conway%27s_Game_of_Life

/// шаг инкремента ячеек (увеличивает ячейки с заданным отступом, в зависимости от текущего цвета пикселей)
void conways_life_incstep(int dx, int dy)  // считает суммы построчно для улучшения локальности доступа
{
  // вычислить отступы в cells и pixels в соответствии с dx, dy
  int cells_x, cells_y, pixels_x, pixels_y;
  if (dx < 0)
  {
    cells_x   =   0;
    pixels_x  = -dx;
  }
  else
  {
    cells_x   = dx;
    pixels_x  =  0;
  }

  if (dy < 0)
  {
    cells_y   =   0;
    pixels_y  = -dy;
  }
  else
  {
    cells_y   = dy;
    pixels_y  =  0;
  }

  const Pixel *pixels_line = pixels + pixels_y * width; // стартовая строчка в pixels
  Cell *cells_line = cells + cells_y * width; // стартовая строчка в cells

  const int w = width  - std::abs(dx), // ширина каждой строчки
            h = height - std::abs(dy); // количество строчек

  for (int line = 0; line < h; ++line)
  {
    // увеличиваем ячейку, если соответствующий пиксель -- не чёрный
    for (int col = 0; col < w; ++col)
    {
      if (pixels_line[pixels_x + col].binary != 0)
        cells_line[cells_x + col]++;
    }
    // перейти к следующей строчке
    pixels_line += width;
    cells_line  += width;
  }
}


//
// игра "Жизнь" меняет привязку idle-функции GLUT (вызывается из главного цикла при отсутствии сообщений)

void conways_life_step();
/// инициализация и привязка
void conways_life_start()
{
  life_game_width   = width;
  life_game_height  = height;
  cells = new Cell[life_game_width * life_game_height];
  glutIdleFunc(conways_life_step);
}

/// останов
void conways_life_stop()
{
  glutIdleFunc(nullptr);
  delete[] cells;
}


/// алгоритм выполняет суммирование "живых" соседей ячеек в cells и заполняет pixels соответственно правилам "Жизни"
void conways_life_step()
{
  // проверим, не изменились ли размеры окна (и, соответственно, пОля "Жизни")
  if (life_game_height != height || life_game_width != width)
  {
    conways_life_stop();  // выполним перезапуск
    conways_life_start();
  }

  // инициализировать счётчики
  std::memset(cells, 0, height * width * sizeof(Cell));

  // вычислить суммы, 8 положений соседних клеток == 8 смещений
  conways_life_incstep(-1, -1);
  conways_life_incstep(-1,  0);
  conways_life_incstep(-1, +1);
  conways_life_incstep( 0, -1);
  conways_life_incstep( 0, +1);
  conways_life_incstep(+1, -1);
  conways_life_incstep(+1,  0);
  conways_life_incstep(+1, +1);

  // получить новые значения pixels, в соответствии с правилом "Жизнь", "наивная" реализация
  for (int i = 0, wh = width * height; i < wh; ++i)
  {
    if (pixels[i].binary != 0) // клетка погибает, если имеет < 2 или > 3 соседей
    {
      if (cells[i] < 2 || cells[i] > 3)
        pixels[i].binary = 0;
    } // попробуйте закомментировать первую ветку if (оставив только ветку if (cells[i] == 3))
    else if (cells[i] == 3)  // место с 3 соседними клетками занимается новой клеткой
    {
      pixels[i] = pencil;
    }
  }

  // TASK: реализовать какие-нибудь альтернативные правила
  // см. http://en.wikipedia.org/wiki/Life-like_cellular_automaton

  // показать
  glutPostRedisplay();
}


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  if (!pixels) return;
  glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
  glFlush();
  
  // счётчик кадров в секунду, выводит среднее за 10с
  ++frames;
  const double dt = avo::seconds() - frames_reported_last;
  if (dt >= 10.0)
  {
    std::cout << frames / dt << '\n';
    frames = 0;
    frames_reported_last += dt;
  }
  
  glFinish();
}


/// изменение размеров окна, функция получает "новые размеры"
void reshape(int w, int h)
{
  // получить положение клиентской области окна
  mouse_origin_x = std::max(0, glutGet(GLUT_WINDOW_X));
  mouse_origin_y = std::max(0, glutGet(GLUT_WINDOW_Y));

  // размеры изменились и корректны? (вторая проверка избыточна на практике)
  if ((w != width || h != height) && (w > 0 && h > 0))
  {
    Pixel *new_pixels = black_canvas(w, h);
    if (pixels) // есть, что копировать
    {
      // копируем прямоугольник, вписывающийся в новые координаты (с левого нижнего угла)
      const int min_w = std::min(w, width),
                min_h = std::min(h, height),
                min_line_size = min_w * sizeof(Pixel); // размер копируемой строки в байтах

      for (int line = 0; line < min_h; ++line)
      {
        std::memcpy(new_pixels + line * w,     // место назначения
                    pixels     + line * width, // копируемые данные
                    min_line_size);            // размер блока в байтах
      }
    }

    // заменим старое содержимое pixels на новое
    delete[] pixels;
    pixels = new_pixels;

    // установим новые размеры картинки
    width = w;
    height = h;
    if (mouse_x >= width)
      mouse_x = width - 1;
    if (mouse_y >= height)
      mouse_y = height - 1;
    
    // установить новые координаты окна просмотра (вся клиентская область)
    glViewport(0, 0, w, h);
  } 
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case ' ': 
  case '\r': case '\n':  black_canvas(); break;
  case 'r':  case 'R':   random_dots();  break;
  case 'm':  case 'M':   simple_maze();  break;
  case 'l':  case 'L':
    // стартовать или остановить игру "Жизнь"
    if (life_game_running = !life_game_running) // изменяет значение признака на противоположное
      conways_life_start();
    else
      conways_life_stop();
    break;

  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    return;

  default:
    return; // другие клавиши игнорируются
  }
  glutPostRedisplay();
}


/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_special_key(int key, int x, int y)
{
  // размер строчки пикселей в байтах (используется при прокручивании картинки)
  const int line_size = width * sizeof(Pixel);
  // нажатие на стрелки прокручивает картинку
  switch (key)
  {
  case GLUT_KEY_LEFT:
    for (int line = 0; line < height; ++line)
    {
      Pixel* base = pixels + line * width;
      Pixel stored = base[0];
      std::memmove(base, base + 1, line_size - sizeof(Pixel));
      base[width - 1] = stored;
    }
    break;

  case GLUT_KEY_UP:
    // используем дополнительную строчку в конце массива pixels
    {
      Pixel* extra_line = pixels + width * height;
      std::memcpy(extra_line, extra_line - width, line_size);
      std::memmove(pixels + width, pixels, line_size * (height - 1));
      std::memcpy(pixels, extra_line, line_size);
    }
    break;

  case GLUT_KEY_RIGHT:
    for (int line = 0; line < height; ++line)
    {
      Pixel* base = pixels + line * width;
      Pixel stored = base[width - 1];
      std::memmove(base + 1, base, line_size - sizeof(Pixel));
      base[0] = stored;
    }
    break;

  case GLUT_KEY_DOWN:
    // используем дополнительную строчку в конце массива pixels
    {
      Pixel* extra_line = pixels + width * height;
      std::memcpy(extra_line, pixels, line_size);
      std::memmove(pixels, pixels + width, line_size * (height - 1));
      std::memcpy(extra_line - width, extra_line, line_size);
    }
    break;

  default:
    return; // другие клавиши игнорируются
  }
  glutPostRedisplay();
}



// проверка качества координат мыши
inline bool good_coords(int x, int y)
{
  return x >= 0 && y >= 0 && x < width && y < height;
}


/// клик мыши, рисуем точку по левому клику
void mouse_click(int button, int state, int x, int y)
{
  y = height - y; // координаты мыши отсчитываются с верхнего левого угла
  if (button != GLUT_LEFT_BUTTON || !good_coords(x, y))
    return;

  if (draw = state == GLUT_DOWN) // нажали?
  {
    // сохранить текущие координаты (для отрисовки линии)
    mouse_x = x;
    mouse_y = y;
    // изобразить точку
    pixels[x + y * width] = pencil;
    glutPostRedisplay();
  }
}


/// отработать движение мыши (рисует отрезок (mouse_x, mouse_y) -- (x, y), если зажата левая кнопка)
void mouse_move(int x, int y)
{
  y = height - y; // координаты мыши отсчитываются с верхнего левого угла
  if (draw && good_coords(x, y))
  {
    line(mouse_x, mouse_y, x, y);
    mouse_x = x;
    mouse_y = y;
    glutPostRedisplay();
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);
  
  // инициализировать pixels на случай невызова reshape на старте (так происходит в сборке MSVC2010)
  black_canvas();

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_SINGLE); // режим: 32-битный цвет RGBA, одинарный буфер кадра SINGLE
  glutInitWindowPosition(0, 0);       // координаты верхнего левого угла окна на экране
  glutInitWindowSize(width, height);  // полные ширина и высота окна
  glutCreateWindow("Pencil");         // создать окно, в функцию передаётся заголовок создаваемого окна

  // привязка функций-обработчиков событий
  glutDisplayFunc(repaint);             // перерисовка
  glutReshapeFunc(reshape);             // изменение размеров окна
  glutKeyboardFunc(handle_key);         // обработка нажатий "обычных" клавиш
  glutSpecialFunc(handle_special_key);  // обработка нажатий "специальных" клавиш
  glutMouseFunc(mouse_click);           // обработка нажатий кнопок мыши
  glutMotionFunc(mouse_move);           // обработка перемещения мыши
  // сменить указатель мыши на крест
  glutSetCursor(GLUT_CURSOR_CROSSHAIR);

  // старт
  frames_reported_last = avo::seconds();
  std::cout << "Starting Pencil. FPS:\n";
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
