﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: Snow2D
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Рисует падающий цветной "снег".
 * Движение всех "снежинок" вычисляется одинаково, всё различие между ними состоит лишь
 * в начальной точке траектории и времени начала падения (ну и цвет разный).
 * Движение состоит из двух фаз: "раскачивающееся" и "прямолинейное".
 * Фактически движение частицы (снежинки) управляется конечным автоматом, который может находиться
 * в трёх состояниях: 
 *   1) раскачивание влево (LeftWave), 
 *   2) раскачивание вправо (RightWave) и
 *   3) прямолинейное падение (Down).
 * Переход на прямолинейное падение происходит при достижении нижней (1/10) части окна.
 * Раскачивание представляет собой движение по ветви параболы.
 * Состояние частицы включает в себя также "стартовые" (для текущего состояния) время и позицию.
 *
 * Нажатие ENTER/SPACE генерирует новый набор псевдослучайных снежинок.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cstdlib>   // rand/srand
#include <iostream>

#include <avo/Timer.hpp>
#include <avo/Color.hpp> // RGBA-цвет в 128-битном формате (компоненты типа float)

using avo::Color;

//
// вспомогательные функции
//

/// следующее псевдослучайное float из отрезка [minv, maxv]
inline float random(float minv, float maxv)
{
  return minv + (maxv - minv) * (std::rand() / float(RAND_MAX));
}

/// возведение в квадрат (иногда удобно иметь такую функцию, да)
inline float square(float x)
{
  return x * x;
}

/// сгенерировать некоторый случайный цвет
inline Color random_bright_color()
{
  // три компонента цвета: "яркий", "возможно тусклый", "любой"
  const float a = random(0.85f, 1.0f),
              b = random(0.30f, 1.0f),
              c = random(0.00f, 1.0f);

  // назначим эти компоненты в произвольном порядке компонентам RGB
  // 0: abc, 1: acb, 2: bac, 3: bca, 4: cab, 5: cba
  switch (std::rand() % 6)
  {
  case 0: return Color(a, b, c);
  case 1: return Color(a, c, b);
  case 2: return Color(b, a, c);
  case 3: return Color(b, c, a);
  case 4: return Color(c, a, b);
  case 5: return Color(c, b, a);
  default: return avo::WHITE; /* добавлено, чтобы успокоить компилятор :) */
  }
}


//
// Реализация снежинки.
//

/// снежинка
struct Snowflake
{
  Color color;      // цвет снежинки
  float t0, x0, y0; // начальные (для текущего состояния) время и позиция снежинки
  // В общем случае не следует хранить абсолютные отметки времени во float, так как его диапазон довольно мал.
  // Однако, в нашем конкретном случае необходимости в double не возникает.
  enum Mode { LeftWave, RightWave, Down } mode; // текущее состояние (режим движения)

  /// установить случайные начальное время, начальное положение и цвет
  void make_random(float t0min, float t0max) // начальное время выбирается из отрезка [t0min, t0max]
  {
    color = random_bright_color();
    t0 = random(t0min, t0max);
    x0 = random(0.0f, 1.0f);   // положение по горизонтали покрывает всю ширину окна
    y0 = random(1.0f, 1.4f);   // положение по вертикали всегда "над" окном в невидимой области (до 40% высоты окна)
    mode = std::rand() & 1? LeftWave: RightWave; // с равной вероятностью начинать раскачиваться влево или вправо
  }

  /**
   * Передаём текущее время, получаем новые координаты снежинки (по ссылкам x, y).
   * Функция возвращает "false", если снежинку следует удалить (true == "пусть ещё поживёт").
   * Состояние снежинки изменяется соответственно (если t достаточно велико для переключения или y достаточно мало). 
   */
  bool next_step(float t, float &x, float &y)
  {
    const float dt = t - t0;
    if (dt <= 0.0f) // текущее время в прошлом относительно t0 -- стоим на месте
    {
      x = x0;
      y = y0;
      return false;
    }

    // двигаться в соответствии с текущим режимом
    switch (mode)
    {
    case LeftWave: // по параболе влево
      x = x0 - (1.f / 60.f) * dt;
      y = y0 - (1.f / 100.f) * (9.f - square(dt - 3.0f)) * dt;
      if (dt >= 4.0f) // не более 4 секунд
      { // переключиться на движение вправо
        mode = RightWave;
        t0 = t;
        x0 = x;
        y0 = y;
      }
      break;

    case RightWave: // по параболе вправо
      x = x0 + (1.f / 60.f) * dt;
      y = y0 - (1.f / 100.f) * (9.f - square(dt - 3.0f)) * dt;
      if (dt >= 4.0f) // не более 4 секунд
      { // переключиться на движение влево
        mode = LeftWave;
        t0 = t;
        x0 = x;
        y0 = y;
      }
      break;

    case Down: // гиперболически замедляющееся падение по вертикали
      x = x0;
      y = y0 / (0.5f * dt + 1.0f);
      break;
    }

    // переключиться на падение, если достигли нижней 1/10 окна
    if (y <= 0.1f)
    {
      mode = Down;
      t0 = t;
      x0 = x;
      y0 = y;
    }
      
    // удаляем все снежинки, которые достигли нижней 1/1000 окна
    return y <= 1e-3f;

    // TASK: добавить отталкивание/притягивание снежинок от/к указателю мыши.
    // Получение координат мыши см., например, в примере Collision2D.
  }
};


//
// константы и глобальные переменные
// 

/// общее количество активных снежинок
const int N = 20000;
// TASK: добавить возможность изменения числа активных снежинок на экране во время работы программы

/// объекты снежинок
Snowflake flake[N];

/// генерируем случайные снежинки, заполняя массив flake с помощью make_random
void make_random_flakes()
{
  const float t0 = (float)avo::seconds(),
              t1 = t0 + 60.f; // разброс времени старта -- 1 минута с текущего момента
  
  for (int i = 0; i < N; ++i)
    flake[i].make_random(t0, t1);
}


//
// обработчики событий GLUT
//

/// расчёт и перерисовка сцены
void repaint()
{
  // получить текущее время
  const float t = (float)avo::seconds();  

  // очистить окно, закрасив его цветом очистки
  glClear(GL_COLOR_BUFFER_BIT);

  // перемещение и отрисовка снежинок
  float x, y;
  glBegin(GL_POINTS); // рисуем набор точек
  for (auto &f : flake)
  {
    if (f.next_step(t, x, y))
    {
      // заменим на новую снежинку
      f.make_random(t, t + 2.f);
    }
    else
    {
      // нарисуем
      f.color.select(); // выбрать текущий цвет -- цвет снежинки
      glVertex2f(x, y); // нарисовать точку в координатах снежинки
    }
  }
  glEnd();

  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}


/// вызывается из главного цикла, когда нет событий для обработки
void idle_proc()
{
  // "спать" 8 миллисекунд, ограничивая таким образом перерисовку 1000/8 = 125 кадрами в секунду
  avo::sleep(0.008); // на деле это может быть далеко не 8мс, а как захочет ОС :)
  glutPostRedisplay(); // попросим GLUT выполнить перерисовку
}


/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void handle_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case ' ': case '\r': case '\n':
    make_random_flakes();
    glutPostRedisplay();
    break;

  default:
    return; // остальные коды игнорируются
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowPosition(0, 0); // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600); // полные ширина и высота окна
  glutCreateWindow("Snow 2D");  // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor (через Color::select)
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f); // установить цвет очистки (фон) -- чёрный
  glPointSize(5.f); // установить диаметр точки ("снежинки") -- 5 пикселей
  glEnable(GL_POINT_SMOOTH); // включить "сглаживание точек"
    // -- без этой команды на некоторых устройствах/версиях драйверов точки будут отображаться не кружками, а квадратами

  // 2D "камера"
  glMatrixMode(GL_PROJECTION); // режим матрицы проекции
  glLoadIdentity();            // сбросить в единичную
  glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0); // домножить на матрицу изометрической проекции,
  // координаты в окне: (0, 0) -- нижний левый угол, (1, 1) -- верхний правый угол
  glMatrixMode(GL_MODELVIEW);  // режим матрицы модели
  
  // привязка функций-обработчиков событий
  glutIdleFunc(idle_proc);      // вызывать, когда нет событий
  glutDisplayFunc(repaint);     // вызывать для отрисовки кадра
  glutKeyboardFunc(handle_key); // вызывать для обработки нажатий "обычных" клавиш

  // старт
  std::cout << "Starting Snow2D\n";
  make_random_flakes();
  glutMainLoop(); // вход в цикл обработки событий
  std::cout << "\nbye!";
  return 0;
}
