﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: SortVis
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Реализации генераторов последовательностей.
 * Разные алгоритмы по-разному "относятся" к разным типам входных последовательностей.
 * Например, если последовательность уже отсортирована, то
 * сортировка пузырьком обгонит пирамидальную сортировку.
 */
#pragma once
#include <cstdlib>
#include <cmath>
#include <algorithm>
#include <functional>


/// следующее псевдослучайное число из отрезка [minv, maxv]
template <class Float> // подходит и float и double, целые -- не подходят
inline Float random(Float minv, Float maxv)
{
  return minv + (maxv - minv) * (std::rand() / Float(RAND_MAX));
}

// "случайная" последовательность
template <class It>
void generateRandom(It begin, It end)
{
  std::generate (begin, end, std::bind(random<float>, 0.02f, 0.98f));
}

// монотонно возрастающая (отсортированная) последовательность
template <class It>
void generateSorted(It begin, It end)
{
  generateRandom (begin, end);
  std::sort (begin, end);
}

// монотонно убывающая ("антиотсортированная") последовательность
template <class It>
void generateAntisorted(It begin, It end)
{
  generateRandom (begin, end);
  typedef typename std::iterator_traits<It>::value_type Value;
  std::sort (begin, end, std::greater<Value>());
}

// заполнить "половинки" с помощью указанных генераторов (обобщённый алгоритм)
template <class It, class BlockGenerator1, class BlockGenerator2>
void generateHalves(It begin, It end, BlockGenerator1 bgen1, BlockGenerator2 bgen2)
{
  using std::distance;
  using std::next;
  const It mid = next(begin, distance(begin, end) / 2);
  bgen1 (begin, mid);
  bgen2 (mid,   end);
}

// "битоническая" последовательность
template <class It>
void generateBitonic(It begin, It end)
{
  generateHalves (begin, end, generateSorted<It>, generateAntisorted<It>);
}

// "антибитоническая" последовательность
template <class It>
void generateAntibitonic(It begin, It end)
{
  generateHalves (begin, end, generateAntisorted<It>, generateSorted<It>);
}

// последовательность блоков (обобщённый алгоритм)
template <class It, class BlockGenerator>
void generateBlocks(It begin, It end, BlockGenerator bgen)
{
  // примерно sqrt(size) блоков по sqrt(size) элементов
  using std::sqrt;
  using std::distance;
  using std::next;
  typedef typename std::iterator_traits<It>::difference_type Diff;
  const Diff size = distance(begin, end),
             block_size = static_cast<Diff>(1. + sqrt(0. + size));
  
  for (Diff i = 0; i < size; i += block_size)
  {
    const It block_end = size < i + block_size? end: next(begin, block_size);
    bgen (begin, block_end);
    begin = block_end;
  }
}

// последовательность отсортированных блоков
template <class It>
void generateSortedBlocks(It begin, It end)
{
  generateBlocks (begin, end, generateSorted<It>);
}

// последовательность "антиотсортированных" блоков
template <class It>
void generateAntisortedBlocks(It begin, It end)
{
  generateBlocks (begin, end, generateAntisorted<It>);
}

// TASK: добавить генератор последовательности с большИм количеством одинаковых значений.
