﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file ConfigFile.hpp
 *
 * Средства для чтения файлов-"словарей", состоящих из строчек вида
 * "ключ" = "значение"
 * // комментарий
 * Нераспознанные строчки игнорируются (refreshDictionary сохраняет их нетронутыми).
 *
 * Кавычки не обязательны, но если их нет, то читается только одно "слово".
 * В начале полагается, что ключи и значения -- строки.
 */
#pragma once
#include <map>
#include <string>
#include <sstream>
#include <type_traits>


/// базовый тип словаря
/**
 * На данный момент это map, который хранит уникальные ключи. 
 * При совпадении ключей, будет сохранено только последнее из значений.
 * При сохранении в файл, содержащий несколько записей с совпадающим ключом,
 * все их значения будут заменены на одно.
 *
 * Пока это является потенциальной проблемой лишь для списка лучших игроков,
 * так как там в качестве ключа используется количество очков.
 */
typedef std::map<std::string, std::string> Dictionary;

/// загрузить словарь из файла
Dictionary loadDictionary(const char *fn);

/// обновить файл словарём
void refreshDictionary(const char *fn, const Dictionary&);


/// вспомогательная функция для перевода значений в строки
template <class T>
std::string toString(T t)
{
  std::ostringstream os;
  os << t;
  return os.str();
}

/// вспомогательная конструкция для разбора строк "на месте"
template <class T>
typename std::decay<T>::type // decay "очищает" тип от const, volatile и &
try_parse(const std::string &input, T &&default_value = T())
{
  std::istringstream parser(input);
  typename std::decay<T>::type result = std::forward<T>(default_value);
  parser >> result;
  return result;
}
/*
 * Функция try_parse может быть записана "проще", как
 template <class T>
 T try_parse(const std::string &input, T default_value = T())
 {
    std::istringstream parser(input);
    T result = default_value;
    parser >> result;
    return result;
 }
 */

/// попытаться извлечь значение по ключу и одновременно разобрать его строковое представление
template <class T>
typename std::decay<T>::type
try_get(const Dictionary &dict, const std::string &key, T &&default_value = T())
{
  auto p = dict.find(key);
  return p != dict.end()? 
    try_parse(p->second, std::forward<T>(default_value))
  : std::forward<T>(default_value);
}
/*
 * Функция try_get может быть записана "проще", как
 template <class T>
 T try_get(const Dictionary &dict, const std::string &key, T default_value = T())
 {
   auto p = dict.find(key);
   return p != dict.end()? try_parse(p->second, default_value) : default_value;
 }
 */
