﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file GameScreen.cpp
 *
 * Реализация GameScreen.hpp.
 */
#include <cstdlib>
#include <vector>
#include <iterator>


#include "GameScreen.hpp"
#include "Game.hpp"
#include "ConfigFile.hpp"

#include <avo/Timer.hpp>
#include <avo/Color.hpp>


using namespace avo;
using namespace std;

//
/// цвета, используемые по умолчанию
namespace colors
{
const Color
  background  (0.2f, 0.6f, 0.0f),
  level       (1.0f, 1.0f, 0.0f),
  score       (1.0f, 1.0f, 1.0f),
  field       (0.4f, 0.7f, 0.2f),

  snake_1     (0.0f, 0.8f, 0.2f),
  snake_2     (0.8f, 0.8f, 1.0f),
  snake_3     (0.5f, 0.5f, 0.7f),
  snake_4     (0.0f, 0.7f, 0.4f),

  rabbit_1_1  (0.4f, 0.0f, 0.0f),
  rabbit_1_2  (0.8f, 0.0f, 0.2f),
  rabbit_1_3  (1.0f, 0.7f, 0.0f),
  rabbit_1_4  (0.5f, 0.3f, 0.0f),

  rabbit_2_1  (0.3f),
  rabbit_2_2  (0.8f),
  rabbit_2_3  (0.4f, 0.4f, 1.0f),
  rabbit_2_4  (0.1f, 0.1f, 0.5f),

  rabbit_3_1  (0.4f, 0.0f, 0.6f),
  rabbit_3_2  (0.8f, 0.0f, 1.0f),
  rabbit_3_3  (0.0f, 0.5f, 1.0f),
  rabbit_3_4  (0.0f, 0.3f, 0.6f);
}



/// чтобы читать цвета, сохранённые в файле, полезно иметь оператор чтения цвета из потока
/**
 * Варианты, количество чисел:
 *  # оттенок серого
 *  # оттенок серого и альфа-канал
 *  # RGB
 *  # RGB и альфа-канал
 *
 */
istream& operator>>(istream &is, Color &color)
{
  float red_or_grey = 0.0f;
  is >> red_or_grey;
  if (!is) return is;

  float green_or_alpha = 0.0f;
  is >> green_or_alpha;
  if (!is) // оттенок серого
  {
    color = Color(red_or_grey);
    if (is.fail()) is.clear ();
    return is;
  }

  float blue = 0.0f;
  is >> blue;
  if (!is) // оттенок серого + прозрачность
  {
    color = Color(red_or_grey, green_or_alpha);
    if (is.fail()) is.clear ();
    return is;
  }

  float alpha = 1.0f;
  is >> alpha;
  if (!is) // RGB
  {
    color = Color(red_or_grey, green_or_alpha, blue);
    if (is.fail()) is.clear();
    return is;
  }

  color = Color(red_or_grey, green_or_alpha, blue, alpha); // RGBA
  return is;
}


/// OpenGL-реализация visitor для Field::scan (рисует поле)
struct CellDrawer
{
  void operator()(int rx, int ry, Cell cell) const
  {
    glTranslatef (dx * rx, dy * ry, 0.0f); // где нарисовать
    glCallList (list_base + cell); // что нарисовать
  }

  /// dx и dy -- шаг по x и по y между ячейками в координатах OpenGL, list_base -- база списков отображения для клеток
  explicit CellDrawer(float dx = 1.0f, float dy = 1.0f, GLuint list_base = 0)
    : dx(dx), dy(dy), list_base(list_base) {}

  /// размеры клетки
  float dx, dy;
  /// база списков отображения (display lists)
  GLuint list_base;
};


/// al'da game
class Game: public ZmeikaGame
{
public:
  Game() { makeLists (); }

  /// отрисовщик игрового поля
  CellDrawer cellDrawer;
  
  /// признак "поставлено на паузу"
  bool paused;

  /// цвет фона
  Color background;
  /// цвет текста -- уровень
  Color level_foreground;
  /// цвет текста -- очки
  Color score_foreground;

  /// начало новой игры
  void start()
  {
    ZmeikaGame::start ();
    paused = true;
    last_time = seconds();
  }

private:

  /// заполнить список отображения для пустого места
  static void makeEmptyCellList(GLuint list, const Color &c)
  {
    glNewList (list, GL_COMPILE);
    glBegin (GL_QUADS);
      c.select ();
      glVertex2f (.05f, .05f); // оставляем зазор, чтобы получить "сеточку" на поле
      glVertex2f (.95f, .05f);
      glVertex2f (.95f, .95f);
      glVertex2f (.05f, .95f);
    glEnd ();
    glEndList ();
  }

  /// заполнить список отображения для квадратика, надо задать 4 цвета
  static void makeBrickList(GLuint list, const Color &c1, const Color &c2, const Color &c3, const Color &c4)
  {
    glNewList (list, GL_COMPILE);
    glBegin (GL_TRIANGLES);
      c1.select ();
        glVertex2f (0.0f, 0.0f);
        glVertex2f (1.0f, 0.0f);
        glVertex2f (0.0f, 1.0f);

      c2.select ();
        glVertex2f (0.0f, 1.0f);
        glVertex2f (1.0f, 0.0f);
        glVertex2f (1.0f, 1.0f);

      c3.select ();
        glVertex2f (0.25f, 0.25f);
        glVertex2f (0.75f, 0.25f);
        glVertex2f (0.25f, 0.75f);

      c4.select ();
        glVertex2f (0.25f, 0.75f);
        glVertex2f (0.75f, 0.25f);
        glVertex2f (0.75f, 0.75f);
    glEnd ();
    glEndList ();
  }

  /// сформировать списки отображения, используемые cellDrawer для отрисовки клеток игрового поля
  void makeLists()
  {
    // прочитать описания цветов
    Dictionary colors = loadDictionary("colors.cfg");
    background = try_get(colors, "background", colors::background);
    level_foreground = try_get(colors, "level", colors::level);
    score_foreground = try_get(colors, "score", colors::score);

    const GLuint lb = glGenLists (Cell_Types); // empty, snake + 3 rabbits
    makeEmptyCellList (lb, try_get(colors, "field", colors::field));
    makeBrickList (lb + 1, 
      try_get(colors, "snake_1", colors::snake_1),
      try_get(colors, "snake_2", colors::snake_2), 
      try_get(colors, "snake_3", colors::snake_3),
      try_get(colors, "snake_4", colors::snake_4));
    makeBrickList (lb + 2,
      try_get(colors, "rabbit_1_1", colors::rabbit_1_1), 
      try_get(colors, "rabbit_1_2", colors::rabbit_1_2),
      try_get(colors, "rabbit_1_3", colors::rabbit_1_3),
      try_get(colors, "rabbit_1_4", colors::rabbit_1_4));
    makeBrickList (lb + 3, 
      try_get(colors, "rabbit_2_1", colors::rabbit_2_1),
      try_get(colors, "rabbit_2_2", colors::rabbit_2_2),
      try_get(colors, "rabbit_2_3", colors::rabbit_2_3),
      try_get(colors, "rabbit_2_4", colors::rabbit_2_4));
    makeBrickList (lb + 4,
      try_get(colors, "rabbit_3_1", colors::rabbit_3_1),
      try_get(colors, "rabbit_3_2", colors::rabbit_3_2),
      try_get(colors, "rabbit_3_3", colors::rabbit_3_3),
      try_get(colors, "rabbit_3_4", colors::rabbit_3_4));
    
    cellDrawer.list_base = lb;
  }
};


//
// GameScreen implementation

GameScreen::GameScreen()
  : impl(nullptr)
{
  impl = new Game;
}

GameScreen::~GameScreen()
{
  delete impl;
}

bool GameScreen::isPaused() const
{
  return impl->paused;
}


void GameScreen::onSelect()
{
  glDisable (GL_BLEND);
  glEnable (GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor (через Color::select)
  glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glLineWidth (2.0f);
  glEnable (GL_LINE_SMOOTH); // включить "сглаживание линий"

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluOrtho2D (0.0, 1.02, 0.0, 1.11);
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  // загрузить описания уровней, начать игру
  impl->start ();

  // установить размеры и позицию окна (размеры поля заданы предыдущим вызовом start)
  fitWindow ((1.02f * 50.f) * impl->field.xsize(), (1.11f * 50.f) * impl->field.ysize());

  // цвет фона (используется при вызове glClear)
  glClearColor (impl->background.red,
                impl->background.green,
                impl->background.blue, 0.0f);
}


void GameScreen::onIdle()
{
  if (!impl->paused)
  {
    impl->paused = impl->step(seconds());
    glutPostRedisplay (); 
  }
  else
  {
    impl->last_time = seconds();
    sleep (0.02); // чтобы не грузил процессор на паузе
  }
}


void GameScreen::onDisplay()
{
  // поле укладывается в "квадрат" [0, 1] x [0, 1] в координатах проекции OpenGL
  glScalef (1.0f / impl->field.xsize(), 1.0f / impl->field.ysize(), 1.0f);
  impl->field.scan (impl->cellDrawer);
  // scan изменяет в процессе текущую матрицу modelview, поэтому сбросим её
  glLoadIdentity ();
  // текст: текущий уровень и количество очков
  impl->level_foreground.select ();
  drawString (GLUT_STROKE_MONO_ROMAN, impl->level.str(),
      0.01f, 1.01f, 0.16f, 1.1f);
  impl->score_foreground.select ();
  drawString (GLUT_STROKE_MONO_ROMAN, impl->score.str(),
      0.19f, 1.01f, 0.19f + 0.1f * impl->score.str().size(), 1.1f);
}


/*
 * W, A, S, D и стрелки -- движение (вверх, влево, вниз и вправо),
 * SPACE/ENTER продолжить/пауза,
 * ESC прекратить игру.
 */

void GameScreen::onChar(unsigned char code, int mouse_x, int mouse_y)
{
  switch (code)
  {
  case 'w': case 'W':  impl->up ();    break;
  case 's': case 'S':  impl->down ();  break;
  case 'a': case 'A':  impl->left ();  break;
  case 'd': case 'D':  impl->right (); break;
  case ' ': case '\n': case '\r':
    impl->paused = !impl->paused;
    break;
  case 'r': case 'R':  // перезапуск игры
    impl->start ();
    break;
  case 27: // перейти к экрану №0 ("список лучших")
    doGameOver (impl->score.getMax()); 
    return;
  default: return;
  }
  glutPostRedisplay ();
}


void GameScreen::onSpecialKey(int code, int mouse_x, int mouse_y)
{
  switch (code)
  {
  case GLUT_KEY_UP:    impl->up ();    break;
  case GLUT_KEY_DOWN:  impl->down ();  break;
  case GLUT_KEY_LEFT:  impl->left ();  break;
  case GLUT_KEY_RIGHT: impl->right (); break;
  default: return;
  }
  glutPostRedisplay ();
}

// телепорт головы
void GameScreen::onMouse(int button, int mouse_x, int mouse_y)
{
  if (button == GLUT_RIGHT_BUTTON)
  {
    // привести координаты мыши к координатам проекции OpenGL
    const float x = float(mouse_x) / glutGet(GLUT_WINDOW_WIDTH),
                y = 1.1f * (1.0f - float(mouse_y) / glutGet(GLUT_WINDOW_HEIGHT));

    if (0.0f <= x && x < 1.0f && 0.0f <= y && y < 1.0f)
    {
      // перевести координаты проекции к координатам поля, переместить голову
      const Field::Position new_head
        (
          static_cast<int>(x * impl->field.xsize()),
          static_cast<int>(y * impl->field.ysize())
        );

      impl->warpSnake (new_head);
      glutPostRedisplay ();
    }
  }
}
