﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file HallOfFameScreen.cpp
 *
 * Реализация HallOfFameScreen.hpp.
 */
#include <cstdlib>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>

#include "HallOfFameScreen.hpp"
#include "ConfigFile.hpp"

#include <avo/Timer.hpp>
using namespace std;
using namespace avo;


//
// HallOfFame

/// имя файла, в котором хранится список лучших игроков
const char * const Hof_Name = "hof";

/// интерфейс к хранилищу достижений пользователей
class HallOfFame
{
public:

  /// загрузить словарь достижений
  void reload()
  {
    top_scores = loadDictionary(Hof_Name);
    make_top_10 ();
  }

  /// обновить словарь достижений, добавив игрока name, заработавшего sc очков
  void refresh(Score sc, const string &name)
  {
    top_scores.emplace (toString(sc), name);
    refreshDictionary (Hof_Name, top_scores);
    make_top_10 ();
  }

  typedef vector<pair<string, string>> Top10;
  /// итератор хранилища пар "строка-строка" (очки-имя) для 10 лучших игроков
  typedef Top10::const_iterator const_iterator;
  const_iterator begin() const
  {
    return top_10.begin();
  }

  const_iterator end() const
  {
    return top_10.end();
  }

private:

  Dictionary top_scores;
  Top10 top_10;

  /// выбрать 10 лучших игроков из словаря top_scores
  void make_top_10 ()
  {
    typedef pair<Score, string> ScoreName; // будем сортировать по числовым значениям, а не строки,
    vector<ScoreName> toBeSorted(top_scores.size()); // поэтому сначала "перегоним" в специальный массив
    transform (top_scores.cbegin(), top_scores.cend(), toBeSorted.begin(),
      [](const Dictionary::value_type &val)
      {
        return make_pair(try_parse<Score>(val.first), val.second);
      });

    // отсортируем и обрежем, чтобы было не более 10
    if (toBeSorted.size() > 10)
    {
      partial_sort (toBeSorted.begin(), toBeSorted.begin() + 10, toBeSorted.end(), std::greater<ScoreName>());
      toBeSorted.resize (10);
    }
    else
    {
      sort (toBeSorted.begin(), toBeSorted.end(), std::greater<ScoreName>());
    }

    // "перегоним" полученный набор обратно в пары (строка, строка)
    top_10.resize (toBeSorted.size());
    transform (toBeSorted.begin(), toBeSorted.end(), top_10.begin(),
      [](pair<Score, string> &val)
      {
        return make_pair(toString(val.first), move(val.second));
      });
  }
};


//
// HallOfFameScreen

HallOfFameScreen::HallOfFameScreen()
  : impl(nullptr)
{
  impl = new HallOfFame;
}

HallOfFameScreen::~HallOfFameScreen()
{
  delete impl;
}

void HallOfFameScreen::addPlayer(Score score, const string &name)
{
  impl->refresh (score, name);
}


void HallOfFameScreen::onSelect()
{
  Base::onSelect ();
  impl->reload ();
}


void HallOfFameScreen::onDisplay()
{
  Base::onDisplay ();

  // до 10 верхних
  auto player = impl->begin(), end_player = impl->end();
  float top = 1.0f, bottom = 0.91f;
  for (int i = 10; i != 0 && player != end_player; --i, ++player) // наличие счётчика i -- перестраховка
  {
    // фон строки текста -- полупрозрачный прямоугольник
    greyRect(bottom, top);

    // сама строка
    random_color().bright().select ();
    drawString (GLUT_STROKE_MONO_ROMAN, player->first,
          0.01f, bottom, 0.3f, top);
    random_color().bright().select ();
    drawString (GLUT_STROKE_ROMAN, player->second,
          0.31f, bottom, 0.41f + 0.05f * player->second.size(), top);
    
    // получить y-координаты следующей строки
    top    -= 0.1f;
    bottom -= 0.1f;
  }
}


void HallOfFameScreen::onChar(unsigned char code, int, int)
{
  switch (code)
  {
  case 27: glutLeaveMainLoop(); return;
  case ' ': case '\n': case '\r': // начать игру
    doGameStart (); break;
  default: ; // ignored
  }
}
