﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Screen.cpp
 *
 * Реализация элементов Screen.h.
 */
#include <GL/freeglut.h>
#include "Screen.hpp"
#include <algorithm>

//
// рисование текста

void drawString(void *font, const std::string &text,
                float left, float bottom, float right, float top)
{
  // функции, вычисляющие "размеры" (в пикселях) не задействуют текущую матрицу,
  // а просто берут значения из зашитых внутри FreeGlut массивов
  auto cstr = reinterpret_cast<const unsigned char*>(text.c_str());
  const GLfloat len  = static_cast<GLfloat>(glutStrokeLength(font, cstr)),
                hght = glutStrokeHeight(font) * (1 + std::count(text.begin(), text.end(), '\n'));

  // теперь желательно вписать эти размеры в заданный прямоугольник
  glPushMatrix ();
  glTranslatef (left, bottom, 0.0f);
  const GLfloat scale_x = left < right? (right - left) / len  : 1.0f,
                scale_y = bottom < top? (top - bottom) / hght : 1.0f;

  if (scale_x != 1.0f || scale_y != 1.0f)
    glScalef (scale_x, scale_y, 1.0f);

  // отрисовка
  glutStrokeString (font, cstr);
  // восстановление текущей матрицы
  glPopMatrix ();
}

//
// подгонка размеров окна

void fitWindow(float target_width, float target_height)
{
  const float scr_width   = static_cast<float>(glutGet(GLUT_SCREEN_WIDTH)),
              scr_height  = glutGet(GLUT_SCREEN_HEIGHT) * 0.9f; // магическое число-поправка, рассчитана на горизонтальную панель задач
                                                                // наверняка где-нибудь будет мешать...

  if (scr_width < target_width)
    target_height *= scr_width / target_width, target_width = scr_width;
  if (scr_height < target_height)
    target_width *= scr_height / target_height, target_height = scr_height;

  glutPositionWindow (static_cast<int>(0.5f * (scr_width - target_width)),
                      static_cast<int>(0.5f * (scr_height - target_height)));
  glutReshapeWindow (static_cast<int>(target_width), static_cast<int>(target_height));
}


//
// AScreen

namespace
{

// текущий экран -- эта глобальная переменная и есть причина невозможности
// переключение экранов из разных параллельно исполняющихся потоков
AScreen *curScreen;

}


AScreen::~AScreen()
{
  if (curScreen == this)
    curScreenSelect (nullptr);
}


void AScreen::select()
{
  if (curScreen != this)
  {
    // отключить
    if (curScreen)         
      curScreen->onDeselect();

    curScreenSelect (this);
  }
}


void AScreen::curScreenSelect(AScreen *newScreen)
{
  if (newScreen == curScreen)
    return;

  if (!newScreen) // => curScreen != nullptr
  {
    // отвязать обработчики
    glutIdleFunc          (nullptr);
    glutDisplayFunc       (nullptr);
    glutReshapeFunc       (nullptr);
    glutVisibilityFunc    (nullptr);
    glutKeyboardFunc      (nullptr);
    glutKeyboardUpFunc    (nullptr);
    glutSpecialFunc       (nullptr);
    glutSpecialUpFunc     (nullptr);
    glutEntryFunc         (nullptr);
    glutPassiveMotionFunc (nullptr);
    glutMotionFunc        (nullptr);
    glutMouseFunc         (nullptr);
    glutMouseWheelFunc    (nullptr);
  }
  else if (!curScreen) // => newScreen != nullptr
  {
    // привязать обработчики
    glutIdleFunc          (&AScreen::onIdleWrapper);
    glutDisplayFunc       (&AScreen::onDisplayWrapper);
    glutReshapeFunc       (&AScreen::onReshapeWrapper);
    glutVisibilityFunc    (&AScreen::onVisibilityWrapper);
    glutKeyboardFunc      (&AScreen::onCharWrapper);
    glutKeyboardUpFunc    (&AScreen::onCharUpWrapper);
    glutSpecialFunc       (&AScreen::onSpecialKeyWrapper);
    glutSpecialUpFunc     (&AScreen::onSpecialKeyUpWrapper);
    glutEntryFunc         (&AScreen::onEntryWrapper);
    glutPassiveMotionFunc (&AScreen::onHoverWrapper);
    glutMotionFunc        (&AScreen::onDragWrapper);
    glutMouseFunc         (&AScreen::onMouseButtonWrapper);
    glutMouseWheelFunc    (&AScreen::onMouseWheelWrapper);
  }

  curScreen = newScreen;
  // подключить
  if (curScreen)
  {
    curScreen->onSelect();
    glutShowWindow ();
  }
}

// готовые обработчики

void AScreen::onReshape(int width, int height)
{
  glViewport (0, 0, width, height);
}

void AScreen::onDisplayBefore()
{
  glClear (GL_COLOR_BUFFER_BIT); // сброс буфера цвета
}

void AScreen::onDisplayAfter()
{
  glutSwapBuffers (); // обмен буфера
}

// обёртки обработчиков событий -- в основном занимаются простой переадресацией

void AScreen::onIdleWrapper()
{
  curScreen->onIdle ();
}

void AScreen::onDisplayWrapper()
{
  curScreen->onDisplayBefore ();
  curScreen->onDisplay ();
  glFinish ();
  curScreen->onDisplayAfter ();
}

void AScreen::onReshapeWrapper(int width, int height)
{
  curScreen->onReshape (width, height);
}

void AScreen::onVisibilityWrapper(int status)
{
  curScreen->onVisibility (status == GLUT_VISIBLE);
}

void AScreen::onCharWrapper(unsigned char code, int mouse_x, int mouse_y)
{
  curScreen->onChar (code, mouse_x, mouse_y);
}

void AScreen::onCharUpWrapper(unsigned char code, int mouse_x, int mouse_y)
{
  curScreen->onCharUp (code, mouse_x, mouse_y);
}

void AScreen::onSpecialKeyWrapper(int code, int mouse_x, int mouse_y)
{
  curScreen->onSpecialKey (code, mouse_x, mouse_y);
}

void AScreen::onSpecialKeyUpWrapper(int code, int mouse_x, int mouse_y)
{
  curScreen->onSpecialKeyUp (code, mouse_x, mouse_y);
}

void AScreen::onEntryWrapper(int status)
{
  curScreen->onEntry (status == GLUT_ENTERED);
}

void AScreen::onHoverWrapper(int mouse_x, int mouse_y)
{
  curScreen->onHover (mouse_x, mouse_y);
}

void AScreen::onDragWrapper(int mouse_x, int mouse_y)
{
  curScreen->onDrag (mouse_x, mouse_y);
}

void AScreen::onMouseButtonWrapper(int button, int state, int x, int y)
{
  if (state == GLUT_DOWN)
    curScreen->onMouse (button, x, y);
  else
    curScreen->onMouseUp (button, x, y);
}

void AScreen::onMouseWheelWrapper(int wheel, int direction, int x, int y)
{
  curScreen->onMouseWheel (wheel, direction, x, y);
}
