﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Screen.hpp
 *
 * Средства рисования.
 */
#pragma once

#include "ViewModelBase.hpp"
#include <string>


/// FreeGlut-реализация средств отображения текста, пытается "впихнуть" строку в указанный прямоугольник
/**
 * @param font может быть GLUT_STROKE_ROMAN или GLUT_STROKE_MONO_ROMAN
 * @param right может быть <= left, тогда выбирается автоматически
 * @param top может быть <= bottom, тогда выбирается автоматически
 * Для выбора толщины линий, которыми рисуется текст, можно использовать glLineWidth(толщина в пикселях).
 */
void drawString(void *font, const std::string &text,
                float left, float bottom, float right = 0.f, float top = 0.f);

/// установить размеры окна (в пикселях)
/** 
  * В случае выхода за пределы экрана, вписывает в экран в пропорции target_width : target_height.
  * Естественно, оба параметра должны быть больше нуля.
  * Позиционирует окно в центре экрана.
  */
void fitWindow(float target_width, float target_height);


/// абстракция текущего экрана + инкапсуляция привязки обработчиков событий FreeGlut
/**
 * Предполагается, что классы-наследники дадут свои определения нужным им обработчикам событий.
 * Готовые элементы реализации предполагают выключенный буфер глубины и включенную двойную буферизацию.
 */
class AScreen
{
public:
  /// базовый деструктор выполняет также "отвязку" данного экрана, если он был активным, но onDeselect уже не вызывает
  virtual ~AScreen();

  /// использовать данный экран для обработки событий (активный экран)
  /**
   * В каждый конкретный момент может быть выбран только один активный экран.
   * Операция не является безопасной при параллельном вызове.
   * Во многопоточном приложении только один поток должен заниматься переключением экранов.
   */
  void select();

protected:
  /// вызывается при выборе данного экрана
  virtual void onSelect() {}
  /// вызывается при выборе другого экрана, если этот экран был активным
  virtual void onDeselect() {}
  /// отсутствие событий, вызывается периодически
  virtual void onIdle() {}
  /// перерисовка кадра
  virtual void onDisplay() = 0;
  /// действие, выполняемое перед отрисовкой кадра
  virtual void onDisplayBefore();
  /// действие, выполняемое после отрисовки кадра
  virtual void onDisplayAfter();
  /// изменение размеров окна
  virtual void onReshape(int width, int height);
  /// изменение видимости окна (присутствует на экране, не присутствует на экране)
  virtual void onVisibility(bool visible) {}
  /// нажатие клавиши (код клавиши преобразуется в соответствующий ASCII-код)
  virtual void onChar(unsigned char code, int mouse_x, int mouse_y) {}
  /// отпускание зажатой клавиши (код клавиши преобразуется в соответствующий ASCII-код)
  virtual void onCharUp(unsigned char code, int mouse_x, int mouse_y) {}
  /// нажатие специальной клавиши
  virtual void onSpecialKey(int code, int mouse_x, int mouse_y) {}
  /// отпускание зажатой специальной клавиши
  virtual void onSpecialKeyUp(int code, int mouse_x, int mouse_y) {}
  /// вход/выход мыши в окно
  virtual void onEntry(bool entered) {}
  /// движение мыши с отпущенными кнопками
  virtual void onHover(int mouse_x, int mouse_y) {}
  /// движение мыши с зажатыми кнопками
  virtual void onDrag(int mouse_x, int mouse_y) {}
  /// нажатие кнопки мыши
  virtual void onMouse(int button, int mouse_x, int mouse_y) {}
  /// отпускание кнопки мыши
  virtual void onMouseUp(int button, int mouse_x, int mouse_y) {}
  /// прокручивание колеса мыши
  /**
   * @param wheel_number -- "номер колеса" (можно игнорировать?)
   * @param direction -- направление прокрутки (теоретически должно быть равно либо +1, либо -1)
   */
  virtual void onMouseWheel(int wheel_number, int direction, int mouse_x, int mouse_y) {}

private:
  // FreeGlut вызывает нижеперечисленные функции в качестве обработчиков событий.
  // Static-функции по сути являются "свободными", и могут быть переданы в качестве колбеков FreeGlut'у.
  // То, что они являются членами AScreen, позволяет им переадресовать вызов
  // обработчику сообщения, являющимуся методом объекта AScreen (вызвать protected-функцию).
  static void curScreenSelect(AScreen*);
  static void onIdleWrapper();
  static void onDisplayWrapper();
  static void onReshapeWrapper(int width, int height);
  static void onVisibilityWrapper(int status);
  static void onCharWrapper(unsigned char code, int mouse_x, int mouse_y);
  static void onCharUpWrapper(unsigned char code, int mouse_x, int mouse_y);
  static void onSpecialKeyWrapper(int code, int mouse_x, int mouse_y);
  static void onSpecialKeyUpWrapper(int code, int mouse_x, int mouse_y);
  static void onEntryWrapper(int status);
  static void onHoverWrapper(int mouse_x, int mouse_y);
  static void onDragWrapper(int mouse_x, int mouse_y);
  static void onMouseButtonWrapper(int button, int state, int x, int y);
  static void onMouseWheelWrapper(int wheel, int direction, int x, int y);
};
